import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/into.js
"""

add = R.add
def isOdd(b): return b % 2 != 0


addXf = {
    '@@transducer/step': add,
    '@@transducer/init': R.always(0),
    '@@transducer/result': R.identity
}


class TestInto(unittest.TestCase):
  def test_transduces_into_arrays(self):
    self.assertEqual([2, 3, 4, 5], R.into([], R.map(add(1)), [1, 2, 3, 4]))
    self.assertEqual([1, 3], R.into([], R.filter(isOdd), [1, 2, 3, 4]))
    # TODO: take

  def test_transduces_into_strings(self):
    self.assertEqual('2345', R.into('', R.map(add(1)), [1, 2, 3, 4]))
    self.assertEqual('13', R.into('', R.filter(isOdd), [1, 2, 3, 4]))
    # TODO: take

  def test_transduces_into_dicts(self):
    self.assertEqual({'a': 1, 'b': 2}, R.into({}, R.identity, [['a', 1], ['b', 2]]))
    self.assertEqual({'a': 1, 'b': 2, 'c': 3}, R.into({}, R.identity, [{'a': 1}, {'b': 2, 'c': 3}]))

  def test_dispatches_to_objects_that_implement_reduce(self):
    pass
    # TODO: we do not support object currently

  def test_allows_custom_transformer(self):
    intoSum = R.into(addXf)
    add2 = R.map(add(2))
    result = intoSum(add2)
    self.assertEqual(18, result([1, 2, 3, 4]))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/values.js
"""

obj = {
    'a': 100,
    'b': [1, 2, 3],
    'c': {
        'x': 200,
        'y': 300
    },
    'd': 'D',
    'e': None
}


class C:
  def __init__(self, a, b):
    self.a = 100
    self.b = 200


cobj = C(100, 200)


class TestValues(unittest.TestCase):
  def test_returns_an_array_of_the_given_object_values(self):
    vs = R.values(obj)
    ts = [100, [1, 2, 3], {'x': 200, 'y': 300}, 'D', None]
    self.assertEqual(len(vs), len(ts))
    self.assertEqual(vs[0], ts[0])
    self.assertEqual(vs[1], ts[1])
    self.assertEqual(vs[2], ts[2])
    self.assertEqual(vs[3], ts[3])
    self.assertEqual(vs[4], ts[4])

  def test_works_on_objects(self):
    self.assertEqual([100, 200], R.values(cobj))

  def test_returns_an_empty_object_for_primitives(self):
    self.assertEqual([], R.values(None))
    self.assertEqual([], R.values(55))
    self.assertEqual([], R.values('foo'))
    self.assertEqual([], R.values(True))
    self.assertEqual([], R.values(False))
    self.assertEqual([], R.values([]))
    self.assertEqual([], R.values({}))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/nth.js
"""

arr = ['foo', 'bar', 'baz', 'quux']


class TestNth(unittest.TestCase):
  def test_accepts_positive_offsets(self):
    self.assertEqual('foo', R.nth(0, arr))
    self.assertEqual('bar', R.nth(1, arr))
    self.assertEqual('baz', R.nth(2, arr))
    self.assertEqual('quux', R.nth(3, arr))
    self.assertEqual(None, R.nth(4, arr))

    self.assertEqual('a', R.nth(0, 'abc'))
    self.assertEqual('b', R.nth(1, 'abc'))
    self.assertEqual('c', R.nth(2, 'abc'))
    self.assertEqual('', R.nth(3, 'abc'))

  def test_accepts_negative_offsets(self):
    self.assertEqual('foo', R.nth(-4, arr))
    self.assertEqual('bar', R.nth(-3, arr))
    self.assertEqual('baz', R.nth(-2, arr))
    self.assertEqual('quux', R.nth(-1, arr))
    self.assertEqual(None, R.nth(-5, arr))

    self.assertEqual('a', R.nth(-3, 'abc'))
    self.assertEqual('b', R.nth(-2, 'abc'))
    self.assertEqual('c', R.nth(-1, 'abc'))
    self.assertEqual('', R.nth(-4, 'abc'))

  def test_throws_if_applied_to_None(self):
    with self.assertRaises(TypeError):
      R.nth(0, None)


if __name__ == '__main__':
  unittest.main()



from .private._curry3 import _curry3
from .private._isFunction import _isFunction
from .private._xReduce import _xReduce
from .private._xwrap import _xwrap


def inner_reduce(xf, acc, arr):
  return _xReduce(_xwrap(xf) if _isFunction(xf) else xf, acc, arr)

reduce = _curry3(inner_reduce)



from math import isnan

from ._helper import getAttribute
from ._isArrayLike import _isArrayLike
from ._isFunction import _isFunction
from ._isNumber import _isNumber


def _equals(a, b):
  if _isNumber(a) and isnan(a) and _isNumber(b) and isnan(b):
    return True
  # pylint: disable=unidiomatic-typecheck
  if type(a) != type(b):
    return False
  if isinstance(a, BaseException):
    # Exception
    return str(a) == str(b)
  if _isArrayLike(a) and _isArrayLike(b):
    # Array-like
    if len(a) != len(b):
      return False
    # pylint: disable=consider-using-enumerate
    for i in range(len(a)):
      if not _equals(a[i], b[i]):
        return False
    return True
  if _isFunction(getAttribute(a, 'equals')) and _isFunction(getAttribute(b, 'equals')):
    # dispatch to objects' own equals method
    return a.equals(b) and b.equals(a)
  # default equals
  return a == b



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/intersection.js
"""

M = [1, 2, 3, 4]
M2 = [1, 2, 3, 4, 1, 2, 3, 4]
N = [3, 4, 5, 6]
N2 = [3, 3, 4, 4, 5, 5, 6, 6]


class TestIntersection(unittest.TestCase):
  def test_combines_two_lists_into_the_set_of_common_elements(self):
    self.assertEqual([3, 4], R.intersection(M, N))

  def test_does_not_allow_duplicates_in_the_output_even_if_the_input_lists_had_duplicates(self):
    self.assertEqual([3, 4], R.intersection(M2, N2))

  def test_does_not_allow_duplicates_in_the_output_even_if_the_first_lists_is_bigger_and_has_duplicates(self):
    self.assertEqual([3, 4], R.intersection(M2, N))

  def test_does_not_allow_duplicates_in_the_output_even_if_the_second_lists_is_bigger_and_has_duplicates(self):
    self.assertEqual([3, 4], R.intersection(M, N2))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simplicity
    self.assertEqual(1, len(R.intersection([float('nan')], [float('nan')])))
    self.assertEqual(1, len(R.intersection([Just([42])], [Just([42])])))


if __name__ == '__main__':
  unittest.main()



from .curryN import curryN
from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._isFunction import _isFunction
from .toString import toString


def inner_invoker(arity, method):
  def wrapper(*args):
    target = args[arity]
    if target is not None and _isFunction(getAttribute(target, method)):
      return getAttribute(target, method)(*args[:arity])
    raise TypeError(f'{toString(target)} does not have method {method}')
  return curryN(arity + 1, wrapper)


invoker = _curry2(inner_invoker)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/unionWith.js
"""

Ro = [{'a': 1}, {'a': 2}, {'a': 3}, {'a': 4}]
So = [{'a': 3}, {'a': 4}, {'a': 5}, {'a': 6}]


def eqA(r, s):
  return r['a'] == s['a']


class TestUnionWith(unittest.TestCase):
  def test_combines_two_lists_into_the_set_of_all_their_elements_based_on_the_passed_in_quiality_predicate(self):
    self.assertEqual([{'a': 1}, {'a': 2}, {'a': 3}, {'a': 4}, {'a': 5}, {'a': 6}], R.unionWith(eqA, Ro, So))


if __name__ == '__main__':
  unittest.main()



from .private._arity import _arity
from .private._curry1 import _curry1
from .private._helper import funcArgsLength


def inner_once(fn):
  called = False
  result = None

  def wrapper(*args):
    nonlocal called, result
    if called:
      return result
    called = True
    result = fn(*args)
    return result
  return _arity(funcArgsLength(fn), wrapper)


once = _curry1(inner_once)



from .private._checkForMethod import _checkForMethod
from .private._curry3 import _curry3


def inner_slice(fromIndex, toIndex, arr):
  if toIndex is None:
    return arr[fromIndex:]
  return arr[fromIndex:toIndex]


# pylint: disable=redefined-builtin
slice = _curry3(_checkForMethod('slice', inner_slice))



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xany import _xany


def inner_any(fn, arr):
  idx = 0
  while idx < len(arr):
    if fn(arr[idx]):
      return True
    idx += 1
  return False


# pylint: disable=redefined-builtin
any = _curry2(_dispatchable(['any'], _xany, inner_any))



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/drop.js
"""


class TestDrop(unittest.TestCase):
  def test_skips_the_first_n_elements_from_a_list_returning_the_remainder(self):
    self.assertEqual(['d', 'e', 'f', 'g'], R.drop(3, ['a', 'b', 'c', 'd', 'e', 'f', 'g']))

  def test_returns_an_empty_array_if_n_is_too_large(self):
    self.assertEqual([], R.drop(20, ['a', 'b', 'c', 'd', 'e', 'f', 'g']))

  def test_returns_an_equivalent_list_if_n_is_less_than_0(self):
    self.assertEqual([1, 2, 3], R.drop(0, [1, 2, 3]))
    self.assertEqual([1, 2, 3], R.drop(-1, [1, 2, 3]))

  def test_never_returns_the_input_array(self):
    xs = [1, 2, 3]
    self.assertEqual(False, xs is R.drop(0, xs))
    self.assertEqual(False, xs is R.drop(-1, xs))

  def test_can_operate_on_strings(self):
    self.assertEqual('da', R.drop(3, 'Ramda'))
    self.assertEqual('a', R.drop(4, 'Ramda'))
    self.assertEqual('', R.drop(5, 'Ramda'))
    self.assertEqual('', R.drop(6, 'Ramda'))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/slice.js
"""


class TestSlice(unittest.TestCase):
  def test_retrieves_the_proper_sublist_of_a_list(self):
    list = [8, 6, 7, 5, 3, 0, 9]
    self.assertEqual([7, 5, 3], R.slice(2, 5, list))

  def test_can_operate_on_tuple(self):
    list = (8, 6, 7, 5, 3, 0, 9)
    self.assertEqual((7, 5, 3), R.slice(2, 5, list))

  def test_can_operate_on_strings(self):
    self.assertEqual(R.slice(0, 0, 'abc'), '')
    self.assertEqual(R.slice(0, 1, 'abc'), 'a')
    self.assertEqual(R.slice(0, 2, 'abc'), 'ab')
    self.assertEqual(R.slice(0, 3, 'abc'), 'abc')
    self.assertEqual(R.slice(0, 4, 'abc'), 'abc')
    self.assertEqual(R.slice(1, 0, 'abc'), '')
    self.assertEqual(R.slice(1, 1, 'abc'), '')
    self.assertEqual(R.slice(1, 2, 'abc'), 'b')
    self.assertEqual(R.slice(1, 3, 'abc'), 'bc')
    self.assertEqual(R.slice(1, 4, 'abc'), 'bc')
    self.assertEqual(R.slice(0, -4, 'abc'), '')
    self.assertEqual(R.slice(0, -3, 'abc'), '')
    self.assertEqual(R.slice(0, -2, 'abc'), 'a')
    self.assertEqual(R.slice(0, -1, 'abc'), 'ab')
    self.assertEqual(R.slice(0, -0, 'abc'), '')
    self.assertEqual(R.slice(-2, -4, 'abc'), '')
    self.assertEqual(R.slice(-2, -3, 'abc'), '')
    self.assertEqual(R.slice(-2, -2, 'abc'), '')
    self.assertEqual(R.slice(-2, -1, 'abc'), 'b')
    self.assertEqual(R.slice(-2, -0, 'abc'), '')

class TestSliceDifferentFromRamda(unittest.TestCase):
  def test_support_None_for_second_param(self):
    self.assertEqual(R.slice(0, None, ['a', 'b', 'c']), ['a', 'b', 'c'])
    self.assertEqual(R.slice(1, None, ['a', 'b', 'c']), ['b', 'c'])
    self.assertEqual(R.slice(2, None, ['a', 'b', 'c']), ['c'])
    self.assertEqual(R.slice(3, None, ['a', 'b', 'c']), [])
    self.assertEqual(R.slice(4, None, ['a', 'b', 'c']), [])

    self.assertEqual(R.slice(0, None, 'abc'), 'abc')
    self.assertEqual(R.slice(1, None, 'abc'), 'bc')
    self.assertEqual(R.slice(2, None, 'abc'), 'c')
    self.assertEqual(R.slice(3, None, 'abc'), '')
    self.assertEqual(R.slice(4, None, 'abc'), '')

if __name__ == '__main__':
  unittest.main()



from .private._arity import _arity
from .private._helper import funcArgsLength
from .private._pipe import _pipe
from .reduce import reduce
from .tail import tail


def pipe(*arguments):
  if len(arguments) == 0:
    raise Exception('pipe requires at least one argument')
  return _arity(
      funcArgsLength(arguments[0]),
      reduce(_pipe, arguments[0], tail(list(arguments)))
  )



from .private._curry1 import _curry1
from .private._has import _has
from .private._isFunction import _isFunction


def inner_keys(obj):
  """
  If the given obj has method keys(), just call it and return key list.
  Otherwise,
  return all attributes defined directly inside of the class, in other words, return obj.__dict__,
  if obj does not has __dict__ attribute, return empty list.

  class A:
    a = 1
    def __init__(self):
      self.b = 2
  class B(A):
    c = 3
    def __init__(self):
      self.d = 4
  b = B()
  keys(b) # ['d'], only attribute d will be included.
  """
  if _has(obj, 'keys') and _isFunction(obj.keys):
    return list(obj.keys())
  if _has(obj, '__dict__'):
    return list(obj.__dict__)
  return []


keys = _curry1(inner_keys)



from .private._curry2 import _curry2
from .private._Set import _Set


def inner_difference(first, second):
  out = []
  idx = 0
  toFilterOut = _Set()
  for item in second:
    toFilterOut.add(item)
  firstLen = len(first)

  while idx < firstLen:
    if toFilterOut.add(first[idx]):
      out.append(first[idx])
    idx += 1

  return out


difference = _curry2(inner_difference)



from .pipe import pipe
from .reverse import reverse


def compose(*arguments):
  if len(arguments) == 0:
    raise Exception('compose requires at least one argument')
  return pipe(*reverse(arguments))



from ._helper import getAttribute
from ._reduced import _reduced
from ._xfBase import XfBase


class XTakeWhile(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f

  def step(self, result, _input):
    if self.f(_input):
      return getAttribute(self.xf, '@@transducer/step')(result, _input)
    return _reduced(result)


def _xtakeWhile(f): return lambda xf: XTakeWhile(f, xf)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/split.js
"""


class TestSplit(unittest.TestCase):
  def test_splits_a_string_into_an_array(self):
    self.assertEqual(['a', 'b', 'c', 'xyz', 'd'], R.split('.', 'a.b.c.xyz.d'))

  def test_the_split_string_can_be_arbitrary(self):
    self.assertEqual(['The C', ' in the H', ' s', ' on the m', ''], R.split('at', 'The Cat in the Hat sat on the mat'))

if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R
from ramda.private._helper import funcArgsLength

"""
https://github.com/ramda/ramda/blob/master/test/flip.js
"""


class TestFlip(unittest.TestCase):
  def test_returns_a_function_which_inverts_the_first_two_arguments_to_the_supplied_function(self):
    def f(a, b, c): return f'{a} {b} {c}'
    g = R.flip(f)
    self.assertEqual('a b c', f('a', 'b', 'c'))
    self.assertEqual('b a c', g('a', 'b', 'c'))

  def test_returns_a_curried_function(self):
    def f(a, b, c): return f'{a} {b} {c}'
    g = R.flip(f)('a')
    self.assertEqual('b a c', g('b', 'c'))

  def test_returns_a_function_with_correct_arity(self):
    def f2(a, b): return f'{a} {b}'
    def f3(a, b, c): return f'{a} {b} {c}'
    self.assertEqual(2, funcArgsLength(R.flip(f2)))
    self.assertEqual(3, funcArgsLength(R.flip(f3)))


class TestFlipProperties(unittest.TestCase):
  def test_inverts_first_two_arguments(self):
    def f(a, b, c): return f'{a} {b} {c}'
    g = R.flip(f)
    self.assertEqual(g('b', 'a', 'c'), f('a', 'b', 'c'))


if __name__ == '__main__':
  unittest.main()



from ._has import _has
from ._isArrayLike import _isArrayLike


def funcArgsLength(fn):
  """
  Get the number of args for function fn
  Not count *args and **kwargs
  """
  return fn.__code__.co_argcount


def toNumber(a):
  """
  Convert any input a to a number type
  if can not convert, then return nan
  """
  if isinstance(a, (int, float)):
    return a
  try:
    return int(a)
  except (ValueError, TypeError):
    try:
      return float(a)
    except (ValueError, TypeError):
      return float('nan')


# pylint: disable=inconsistent-return-statements
def getAttribute(v, key):
  """
  This function is mainly for retrive @@transducer/xxx property, and fantasy-land/xxx property.
  We assume dict/object in Python may own such properties.

  dict case:
    d = {'@@transducer/init': lambda: True}
    init_fn = getAttribute(d, '@@transducer/init')

  obj case:
    class T:
      def init(self):
        return True
      def get(self, type):
        if type == '@@transducer/init':
          return self.init
    t = T()
    init_fn = getAttribute(t, '@@transducer/init')

  method case:
    class Mapper:
      def map(fn):
        return fn
    m = Mapper()
    map_fn = getAttribute(m, 'map')

  return: function got from key, otherwise None
  """
  if isinstance(v, dict) and key in v:
    return v[key]
  if _has(v, key):
    return getattr(v, key, None)
  if _has(v, 'get'):
    try:
      return v.get(key, None)
    except TypeError:
      # in case v has get method but with different signature
      return None


def safeLen(x):
  if _isArrayLike(x):
    return len(x)
  return 0




import unittest
from itertools import count

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/take.js
"""


class TestTake(unittest.TestCase):
  def test_takes_only_the_first_n_elements_from_a_list(self):
    self.assertEqual(['a', 'b', 'c'], R.take(3, ['a', 'b', 'c', 'd', 'e', 'f', 'g']))

  def test_returns_only_as_many_as_the_array_can_provide(self):
    self.assertEqual([1, 2], R.take(3, [1, 2]))
    self.assertEqual([], R.take(3, []))

  def test_returns_an_equivalent_list_if_n_is_negative(self):
    self.assertEqual([1, 2, 3], R.take(-1, [1, 2, 3]))

  def test_never_returns_the_input_array(self):
    xs = [1, 2, 3]
    self.assertIsNot(xs, R.take(3, xs))
    self.assertIsNot(xs, R.take(-1, xs))

  def test_can_operate_on_strings(self):
    self.assertEqual('Ram', R.take(3, 'Ramda'))
    self.assertEqual('Ra', R.take(2, 'Ramda'))
    self.assertEqual('R', R.take(1, 'Ramda'))
    self.assertEqual('', R.take(0, 'Ramda'))

  def test_handles_zero_correctly(self):
    self.assertEqual([], R.into([], R.take(0), [1, 2, 3]))

  def test_steps_correct_number_of_times(self):
    count = 0

    def spy(_):
      nonlocal count
      count += 1

    R.into([], R.compose(R.map(spy), R.take(2)), [1, 2, 3])
    self.assertEqual(2, count)

  def test_transducer_called_for_every_member_of_list_if_n_is_negative(self):
    count = 0

    def spy(_):
      nonlocal count
      count += 1

    R.into([], R.compose(R.map(spy), R.take(-1)), [1, 2, 3])
    self.assertEqual(3, count)


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xtap import _xtap


def inner_tap(fn, x):
  fn(x)
  return x


tap = _curry2(_dispatchable([], _xtap, inner_tap))



from .private._curry1 import _curry1


def inner_comparator(pred):
  def comparator_wrapper(a, b):
    if pred(a, b):
      return -1
    if pred(b, a):
      return 1
    return 0
  return comparator_wrapper


comparator = _curry1(inner_comparator)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/zip.js
"""


class TestZip(unittest.TestCase):
  def test_returns_an_array_of_tuples(self):
    a = [1, 2, 3]
    b = [100, 200, 300]
    self.assertEqual([[1, 100], [2, 200], [3, 300]], R.zip(a, b))

  def test_returns_a_list_as_long_as_the_shorter_of_the_lists_input(self):
    a = [1, 2, 3]
    b = [100, 200, 300, 400]
    c = [10, 20]
    self.assertEqual([[1, 100], [2, 200], [3, 300]], R.zip(a, b))
    self.assertEqual([[1, 10], [2, 20]], R.zip(a, c))

if __name__ == '__main__':
  unittest.main()



from .private._concat import _concat
from .private._curry3 import _curry3


def inner_adjust(idx, fn, arr):
  length = len(arr)
  if idx >= length or idx < -length:
    return arr
  _idx = (length + idx) % length
  _arr = _concat(arr)
  _arr[_idx] = fn(arr[_idx])
  return _arr


adjust = _curry3(inner_adjust)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/sortBy.js
"""

albums = [
    {'title': 'Art of the Fugue', 'artist': 'Glenn Gould', 'genre': 'Baroque'},
    {'title': 'A Farewell to Kings', 'artist': 'Rush', 'genre': 'Rock'},
    {'title': 'Timeout', 'artist': 'Dave Brubeck Quartet', 'genre': 'Jazz'},
    {'title': 'Fly By Night', 'artist': 'Rush', 'genre': 'Rock'},
    {'title': 'Goldberg Variations', 'artist': 'Daniel Barenboim', 'genre': 'Baroque'},
    {'title': 'New World Symphony', 'artist': 'Leonard Bernstein', 'genre': 'Romantic'},
    {'title': 'Romance with the Unseen', 'artist': 'Don Byron', 'genre': 'Jazz'},
    {'title': 'Somewhere In Time', 'artist': 'Iron Maiden', 'genre': 'Metal'},
    {'title': 'In Times of Desparation', 'artist': 'Danny Holt', 'genre': 'Modern'},
    {'title': 'Evita', 'artist': 'Various', 'genre': 'Broadway'},
    {'title': 'Five Leaves Left', 'artist': 'Nick Drake', 'genre': 'Folk'},
    {'title': 'The Magic Flute', 'artist': 'John Eliot Gardiner', 'genre': 'Classical'}
]


class TestSortBy(unittest.TestCase):
  def test_sorts_by_a_simple_property_of_the_objects(self):
    sortedAlbums = R.sortBy(R.prop('title'), albums)
    self.assertEqual(len(albums), len(sortedAlbums))
    self.assertEqual('A Farewell to Kings', sortedAlbums[0]['title'])
    self.assertEqual('Timeout', sortedAlbums[11]['title'])

  def test_preserves_object_identity(self):
    a = {'value': 'a'}
    b = {'value': 'b'}
    result = R.sortBy(R.prop('value'), [a, b])
    self.assertEqual(a, result[0])
    self.assertEqual(b, result[1])

  def test_sorts_array_like_object(self):
    args = (lambda *args: args)('c', 'a', 'b')
    result = R.sortBy(R.identity, args)
    self.assertEqual('a', result[0])
    self.assertEqual('b', result[1])
    self.assertEqual('c', result[2])


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/uniq.js
"""


class TestUniq(unittest.TestCase):
  def test_returns_a_set_from_any_array(self):
    self.assertEqual([1, 2, 3], R.uniq([1, 2, 3, 1, 2, 3, 1, 2, 3]))

  def test_keeps_elements_from_the_left(self):
    self.assertEqual([1, 2, 3, 4], R.uniq([1, 2, 3, 4, 1]))

  def test_returns_an_empty_array_for_an_empty_array(self):
    self.assertEqual([], R.uniq([]))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    # self.assertEqual(2, len(R.uniq([0], [-0])))
    self.assertEqual(1, len(R.uniq([float('nan'), float('nan')])))
    # self.assertEqual(1, len(R.uniq([[1], [1]]))) # python only support hashable obj
    self.assertEqual(1, len(R.uniq([Just([42]), Just([42])])))

  def test_handles_None_elements(self):
    self.assertEqual([None], R.uniq([None, None, None]))

  def test_uses_reference_equality_for_functions(self):
    self.assertEqual(2, len(R.uniq([R.add, R.identity, R.add, R.identity, R.add, R.identity])))


if __name__ == '__main__':
  unittest.main()



from .private._checkForMethod import _checkForMethod
from .private._curry2 import _curry2
from .reduceBy import reduceBy


def inner_groupBy(acc, item):
  acc.append(item)
  return acc


groupBy = _curry2(_checkForMethod('groupBy', reduceBy(inner_groupBy, [])))



import copy

from .keys import keys
from .private._curry2 import _curry2


def inner_pick(names, obj):
  """
  case 1: obj is dict
  case 2: obj is object
  """
  if isinstance(obj, dict):
    res = {}
    for name in names:
      if name in obj:
        res[name] = obj[name]
    return res
  res = copy.copy(obj)
  for key in keys(obj):
    if key not in names:
      delattr(res, key)
  return res


pick = _curry2(inner_pick)



from .private._curry2 import _curry2
from .private._isString import _isString


def inner_nth(offset, arr):
  idx = len(arr) + offset if offset < 0 else offset
  if _isString(arr):
    if 0 <= idx < len(arr):
      return arr[idx]
    return ''
  if 0 <= idx < len(arr):
    return arr[idx]
  return None


nth = _curry2(inner_nth)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/and.js
"""


class TestAnd(unittest.TestCase):
  def test_compares_two_values_with_python_and(self):
    self.assertEqual(True, R.And(True, True))
    self.assertEqual(False, R.And(True, False))
    self.assertEqual(False, R.And(False, True))
    self.assertEqual(False, R.And(False, False))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R
from ramda.private._curry2 import _curry2
from ramda.private._isFunction import _isFunction

from .helpers.listXf import listXf

"""
https://github.com/ramda/ramda/blob/master/test/tap.js
"""

appendToList = _curry2(lambda lst, x: lst.append(x))


class TestTap(unittest.TestCase):
  def test_returns_a_function_that_always_returns_its_argument(self):
    f = R.tap(R.identity)
    self.assertTrue(_isFunction(f))
    self.assertEqual(100, f(100))
    self.assertEqual(None, f(None))

  def test_may_take_a_function_as_the_first_argument_that_executes_with_taps_argument(self):
    sideEffect = 0
    self.assertEqual(0, sideEffect)

    def fn(x):
      nonlocal sideEffect
      sideEffect = 'string ' + str(x)
    rv = R.tap(fn, 200)
    self.assertEqual(200, rv)
    self.assertEqual('string 200', sideEffect)

  def test_can_act_as_a_transducer(self):
    sideEffect = []
    numbers = [1, 2, 3, 4, 5]

    xf = R.compose(R.map(R.identity), R.tap(appendToList(sideEffect)))

    self.assertEqual(numbers, R.into([], xf, numbers))
    self.assertEqual(numbers, sideEffect)

  def test_dispatches_to_transformer_objects(self):
    sideEffect = []
    appendToSideEffect = appendToList(sideEffect)

    res = R.tap(appendToSideEffect, listXf)
    print(res)
    self.assertEqual(appendToSideEffect, res.f)
    self.assertEqual(listXf, res.xf)


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/pickAll.js
"""

obj = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6}


class TestPickAll(unittest.TestCase):
  def test_copies_the_named_properties_of_an_object_to_the_new_object(self):
    self.assertEqual({'a': 1, 'c': 3, 'f': 6}, R.pickAll(['a', 'c', 'f'], obj))

  def test_includes_properties_not_present_on_the_input_object(self):
    self.assertEqual({'a': 1, 'c': 3, 'g': None}, R.pickAll(['a', 'c', 'g'], obj))

  def test_work_with_objects(self):
    class Obj:
      def __init__(self, a, b, c, d, e, f):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
    o = Obj(1, 2, 3, 4, 5, 6)
    omittedObj = R.pick(['a', 'c', 'f'], o)
    self.assertEqual(1, omittedObj.a)
    self.assertEqual(False, hasattr(omittedObj, 'b'))
    self.assertEqual(3, omittedObj.c)
    self.assertEqual(False, hasattr(omittedObj, 'd'))
    self.assertEqual(False, hasattr(omittedObj, 'e'))
    self.assertEqual(6, omittedObj.f)

    # not changed the original object
    self.assertEqual(1, o.a)
    self.assertEqual(2, o.b)
    self.assertEqual(3, o.c)
    self.assertEqual(4, o.d)
    self.assertEqual(5, o.e)
    self.assertEqual(6, o.f)

  def test_work_with_objects_includes_properties_not_present_on_the_input_object(self):
    class Obj:
      def __init__(self, a, b, c, d, e, f):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
    o = Obj(1, 2, 3, 4, 5, 6)
    omittedObj = R.pickAll(['a', 'c', 'g'], o)
    self.assertEqual(1, omittedObj.a)
    self.assertEqual(False, hasattr(omittedObj, 'b'))
    self.assertEqual(3, omittedObj.c)
    self.assertEqual(False, hasattr(omittedObj, 'd'))
    self.assertEqual(False, hasattr(omittedObj, 'e'))
    self.assertEqual(None, omittedObj.g)

    # not changed the original object
    self.assertEqual(1, o.a)
    self.assertEqual(2, o.b)
    self.assertEqual(3, o.c)
    self.assertEqual(4, o.d)
    self.assertEqual(5, o.e)
    self.assertEqual(6, o.f)


if __name__ == '__main__':
  unittest.main()



import ramda as R
from ramda.private._helper import getAttribute

sentinel = {}


class Maybe:
  def __init__(self, x, box):
    if x != sentinel:
      raise Exception('Can not instantiate Maybe')
    self.isJust = len(box) > 0
    if self.isJust:
      self.value = box[0]
    self.isNothing = not self.isJust

  def equals(self, other):
    if other is not None and getAttribute(other, '@@type') == getAttribute(self, '@@type') and self.isJust:
      return other.isJust and R.equals(self.value, other.value)
    else:
      return other.isNothing

  def map(self, f):
    if self.isJust:
      return Just(f(self.value))
    else:
      return Nothing()

  def ap(self, maybe):
    if self.isJust and getAttribute(maybe, 'isJust'):
      return Just(maybe.value(self.value))
    else:
      return Nothing()

  def chain(self, f):
    if self.isJust:
      return f(self.value)
    else:
      return Nothing()

  def filter(self, pred):
    if self.isJust and pred(self.value):
      return self
    else:
      return Nothing()

  def toSring(self):
    if self.isJust:
      return 'Just(' + R.toString(self.value) + ')'
    else:
      return 'Nothing'

  def empty(self):
    return Nothing()

  def get(self, name, default=None):
    if name == 'fantasy-land/of':
      return Just
    if name == '@@type':
      return 'ramda/Maybe'
    if name == 'fantasy-land/equals':
      return self.equals
    if name == 'fantasy-land/map':
      return self.map
    if name == 'fantasy-land/ap':
      return self.ap
    if name == 'fantasy-land/chain':
      return self.chain
    if name == 'fantasy-land/filter':
      return self.filter
    if name == 'fantasy-land/empty':
      return self.empty
    return default


class Nothing(Maybe):
  def __init__(self, value=[]):
    super().__init__(sentinel, value)


class Just(Maybe):
  def __init__(self, value):
    super().__init__(sentinel, [value])



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/prepend.js
"""


class TestPrepend(unittest.TestCase):
  def test_adds_the_element_to_the_beginning_of_the_list(self):
    self.assertEqual(['x', 'y', 'z'], R.prepend('x', ['y', 'z']))
    self.assertEqual([['a', 'z'], 'x', 'y'], R.prepend(['a', 'z'], ['x', 'y']))

  def test_works_on_empty_list(self):
    self.assertEqual([1], R.prepend(1, []))


if __name__ == '__main__':
  unittest.main()



from functools import cmp_to_key

from .private._curry2 import _curry2


def inner_sortBy(fn, arr):
  def comparator(a, b):
    aa = fn(a)
    bb = fn(b)
    if aa < bb:
      return -1
    if aa > bb:
      return 1
    return 0
  return sorted(arr, key=cmp_to_key(comparator))


sortBy = _curry2(inner_sortBy)




import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/lastIndexOf.js
"""

input = [1, 2, 3, 4, 5, 1]


class TestLastIndexOf(unittest.TestCase):
  def test_returns_a_number_indicating_an_object_last_position_in_a_list(self):
    arr = [0, 10, 20, 30, 0, 10, 20, 30, 0, 10]
    self.assertEqual(7, R.lastIndexOf(30, arr))

  def test_returns_minus_1_if_the_object_is_not_in_the_list(self):
    arr = [0, 10, 20, 30]
    self.assertEqual(-1, R.lastIndexOf(40, arr))

  def test_returns_the_last_index_of_the_first_item(self):
    self.assertEqual(5, R.lastIndexOf(1, input))

  def test_returns_the_index_of_the_last_item(self):
    self.assertEqual(4, R.lastIndexOf(5, input))

  def test_does_not_consider_str_1_equals_to_number_1(self):
    self.assertEqual(-1, R.lastIndexOf('1', ['a', 1, 'a']))

  def test_returns_minus_1_for_an_empty_array(self):
    self.assertEqual(-1, R.lastIndexOf('x', []))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    # self.assertEqual(-1, R.lastIndexOf(0, [-0]))
    self.assertEqual(0, R.lastIndexOf(float('nan'), [float('nan')]))
    self.assertEqual(0, R.lastIndexOf(Just([42]), [Just([42])]))

if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/keys.js
"""

obj = {'a': 100, 'b': [1, 2, 3], 'c': {'x': 200, 'y': 300}, 'd': 'D', 'e': None}


class A:
  def __init__(self):
    self.a = 'a'


class B(A):
  def __init__(self):
    self.b = 'b'


class C(A):
  def __init__(self):
    super().__init__()
    self.c = 'c'


class TestKeys(unittest.TestCase):
  def test_returns_an_array_of_the_given_object_own_keys(self):
    self.assertEqual(['a', 'b', 'c', 'd', 'e'], R.keys(obj))

  def test_works_with_attribute_override(self):
    a, b, c = A(), B(), C()
    self.assertEqual(['a'], R.keys(a))
    self.assertEqual(['b'], R.keys(b))
    self.assertEqual(['a', 'c'], R.keys(c))

  def test_works_with_primitives(self):
    self.assertEqual([], R.keys(None))
    self.assertEqual([], R.keys(55))
    self.assertEqual([], R.keys('foo'))
    self.assertEqual([], R.keys(True))
    self.assertEqual([], R.keys(False))
    self.assertEqual([], R.keys([]))
    self.assertEqual([], R.keys({}))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.listXf import listXf

"""
https://github.com/ramda/ramda/blob/master/test/all.js
"""


def even(n): return n % 2 == 0
def T(): return True
def isFalse(x): return x == False


intoArray = R.into([])


class TestAll(unittest.TestCase):
  def test_returns_true_if_all_elements_satisfy_the_predicate(self):
    self.assertEqual(True, R.all(even, [2, 4, 6, 8, 10, 12]))
    self.assertEqual(True, R.all(isFalse, [False, False, False]))

  def test_false_if_any_element_fails_to_satisfy_the_predicate(self):
    self.assertEqual(False, R.all(even, [2, 4, 6, 8, 9, 10]))

  def test_true_for_an_empty_list(self):
    self.assertEqual(True, R.all(T, []))

  def test_true_into_array_if_all_elements_satisfy_the_predicate(self):
    self.assertEqual([True], intoArray(R.all(even), [2, 4, 6, 8, 10, 12]))
    self.assertEqual([True], intoArray(R.all(isFalse), [False, False, False]))

  def test_false_into_array_if_any_element_failes_to_satisfy_the_predicate(self):
    self.assertEqual([False], intoArray(R.all(even), [2, 4, 6, 8, 9, 10]))

  def test_true_into_array_for_an_empty_list(self):
    self.assertEqual([True], intoArray(R.all(T), []))

  def test_works_with_more_complex_objects(self):
    xs = [{'x': 'abc'}, {'x': 'ade'}, {'x': 'fghiajk'}]
    def len3(o): return len(o.get('x')) == 3
    def hasA(o): return 'a' in o.get('x')
    self.assertEqual(False, R.all(len3, xs))
    self.assertEqual(True, R.all(hasA, xs))

  def test_dispatches_when_given_a_transformer_in_list_position(self):
    res = R.all(even, listXf)
    self.assertEqual(res.all, True)
    self.assertEqual(res.f, even)
    self.assertEqual(res.xf, listXf)

  def test_can_act_as_a_transducer(self):
    input = [2, 4, 6, 8, 9, 10]
    expected = [False]
    self.assertEqual(expected, R.into([], R.all(even), input))
    # TODO: R.transduce


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/reject.js
"""


def even(x): return x % 2 == 0 if isinstance(x, int) else False


class FilterObject:
  def filter(self, f):
    return f('called f.filter')


class Obj:
  def fn(self):
    return 'called fn'


class TestReject(unittest.TestCase):
  def test_reduces_an_array_to_those_not_matching_a_filter(self):
    self.assertEqual([1, 3, 5], R.reject(even, [1, 2, 3, 4, 5]))

  def test_reduces_an_empty_array_if_no_element_matches(self):
    self.assertEqual([], R.reject(lambda x: x < 100, [1, 9, 99]))

  def test_reduces_an_empty_array_if_asked_to_filter_an_empty_array(self):
    self.assertEqual([], R.reject(lambda x: x > 100, []))

  def test_filter_dict(self):
    def equals0(x): return x == 0
    self.assertEqual({}, R.reject(equals0, {}))
    self.assertEqual({}, R.reject(equals0, {'x': 0, 'y': 0, 'z': 0}))
    self.assertEqual({'x': 1}, R.reject(equals0, {'x': 1, 'y': 0, 'z': 0}))
    self.assertEqual({'x': 1, 'y': 2}, R.reject(equals0, {'x': 1, 'y': 2, 'z': 0}))
    self.assertEqual({'x': 1, 'y': 2, 'z': 3}, R.reject(equals0, {'x': 1, 'y': 2, 'z': 3}))

  def test_filter_objects(self):
    def equals0(x): return x == 0
    o1 = Obj()
    o1.x = 0
    o1.y = 0
    o1.z = 0
    o1_expected = Obj()
    self.assertEqual(o1_expected.__dict__, R.reject(equals0, o1).__dict__)

    o2 = Obj()
    o2.x = 1
    o2.y = 0
    o2.z = 0
    o2_expected = Obj()
    o2_expected.x = 1
    self.assertEqual(o2_expected.__dict__, R.reject(equals0, o2).__dict__)

    o3 = Obj()
    o3.x = 1
    o3.y = 2
    o3.z = 0
    o3_expected = Obj()
    o3_expected.x = 1
    o3_expected.y = 2
    self.assertEqual(o3_expected.__dict__, R.reject(equals0, o3).__dict__)

    o4 = Obj()
    o4.x = 1
    o4.y = 2
    o4.z = 3
    o4_expected = Obj()
    o4_expected.x = 1
    o4_expected.y = 2
    o4_expected.z = 3
    self.assertEqual(o4_expected.__dict__, R.reject(equals0, o4).__dict__)

  def test_filter_objects_keep_functions(self):
    o = Obj()
    o_filtered = R.reject(lambda x: x > 0, o)
    self.assertEqual('called fn', o_filtered.fn())

  def test_dispatches_to_filter_method(self):
    m = Just(42)
    self.assertEqual(m, R.filter(R.T, m))
    self.assertTrue(R.filter(R.F, m).isNothing)
    self.assertTrue(R.reject(R.T, m).isNothing)
    self.assertEqual(m, R.reject(R.F, m))

  def test_can_act_as_a_transducer(self):
    input = [1, 2, 3, 4]
    expected = [1, 3]
    self.assertEqual(expected, R.into([], R.reject(even), input))
    # TODO: R.transduce


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/zipWith.js
"""

a = [1, 2, 3]
b = [100, 200, 300]
c = [10, 20, 30, 40, 50, 60]
def s(a, b): return f'{a} cow {b}'


class TestZipWith(unittest.TestCase):
  def test_returns_an_array_created_by_applying_its_passed_in_function_pair_wise_on_its_passed_in_arrays(self):
    self.assertEqual([101, 202, 303], R.zipWith(R.add, a, b))
    self.assertEqual([100, 400, 900], R.zipWith(R.multiply, a, b))
    self.assertEqual(['1 cow 100', '2 cow 200', '3 cow 300'], R.zipWith(s, a, b))

  def test_returns_an_array_whose_length_is_equal_to_the_shorter_of_its_input_arrays(self):
    self.assertEqual(len(a), len(R.zipWith(R.add, a, c)))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/takeWhile.js
"""


class TestTakeWhile(unittest.TestCase):
  def test_continues_taking_elements_while_the_function_reports_true(self):
    self.assertEqual([1, 3], R.takeWhile(lambda x: x != 5, [1, 3, 5, 7, 9]))

  def test_starts_at_the_right_arg_and_acknowledges_undefined(self):
    self.assertEqual([], R.takeWhile(lambda: False, []))
    self.assertEqual([1, 3], R.takeWhile(lambda x: x is not None, [1, 3, None, 5, 7]))

  def test_can_operate_on_strings(self):
    self.assertEqual('Ram', R.takeWhile(lambda x: x != 'd', 'Ramda'))

  def test_can_act_as_a_transducer(self):
    def isNotFour(x): return x != 4
    input = [1, 2, 3, 4, 3, 2, 1]
    expected = [1, 2, 3]
    self.assertEqual(expected, R.into([], R.takeWhile(isNotFour), input))
    # TODO: transducer


if __name__ == '__main__':
  unittest.main()



from ._curry1 import _curry1
from ._curry2 import _curry2
from ._isPlaceholder import _isPlaceholder


def _curry3(fn):
  # pylint: disable=too-many-return-statements
  def f3(*args):
    if len(args) >= 1:
      a = args[0]
    if len(args) >= 2:
      b = args[1]
    if len(args) >= 3:
      c = args[2]

    def f_ab(_a, _b): return fn(_a, _b, c)
    def f_bc(_b, _c): return fn(a, _b, _c)
    def f_ac(_a, _c): return fn(_a, b, _c)
    def f_a(_a): return fn(_a, b, c)
    def f_b(_b): return fn(a, _b, c)
    def f_c(_c): return fn(a, b, _c)
    if len(args) == 0:
      return f3
    if len(args) == 1:
      if _isPlaceholder(a):
        return f3
      return _curry2(f_bc)
    if len(args) == 2:
      if _isPlaceholder(a) and _isPlaceholder(b):
        return f3
      if _isPlaceholder(a):
        return _curry2(f_ac)
      if _isPlaceholder(b):
        return _curry2(f_bc)
      return _curry1(f_c)
    if _isPlaceholder(a) and _isPlaceholder(b) and _isPlaceholder(c):
      return f3
    if _isPlaceholder(a) and _isPlaceholder(b):
      return _curry2(f_ab)
    if _isPlaceholder(a) and _isPlaceholder(c):
      return _curry2(f_ac)
    if _isPlaceholder(b) and _isPlaceholder(c):
      return _curry2(f_bc)
    if _isPlaceholder(a):
      return _curry1(f_a)
    if _isPlaceholder(b):
      return _curry1(f_b)
    if _isPlaceholder(c):
      return _curry1(f_c)
    return fn(a, b, c)
  return f3



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/flatten.js
"""


class TestFlatten(unittest.TestCase):
  def test_turns_a_nested_list_into_one_flat_list(self):
    nest = [1, [2], [3, [4, 5], 6, [[[7], 8]]], 9, 10]
    self.assertEqual([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], R.flatten(nest))
    nest = [[[[3]], 2, 1], 0, [[-1, -2], -3]]
    self.assertEqual([3, 2, 1, 0, -1, -2, -3], R.flatten(nest))
    self.assertEqual([1, 2, 3, 4, 5], R.flatten([1, 2, 3, 4, 5]))

  def test_is_not_destructive(self):
    nest = [1, [2], [3, [4, 5], 6, [[[7], 8]]], 9, 10]
    self.assertIsNot(nest, R.flatten(nest))

  # def test_handle_ridiculously_large_inputs(self):
  # TODO: find a good way to test timeout
  #   self.assertEqual(1056003, len(R.flatten([*[1] * 1000000, R.range(0, 56000), 5, 1, 3])))

  def test_handles_array_like_objects(self):
    o = {0: [1, 2, [3]], 1: [], 2: ['a', 'b', 'c', ['d', 'e']]}
    self.assertEqual([1, 2, 3, 'a', 'b', 'c', 'd', 'e'], R.flatten(o))

  def test_flattens_an_array_of_empty_arrays(self):
    self.assertEqual([], R.flatten([[], [], []]))
    self.assertEqual([], R.flatten([]))

if __name__ == '__main__':
  unittest.main()



from ._helper import getAttribute
from ._includesWith import _includesWith
from ._xfBase import XfBase


class XUniqWith(XfBase):
  def __init__(self, pred, xf):
    self.xf = xf
    self.pred = pred
    self.items = []

  def step(self, result, _input):
    if _includesWith(self.pred, _input, self.items):
      return result
    self.items.append(_input)
    return getAttribute(self.xf, '@@transducer/step')(result, _input)


def _xuniqWith(pred): return lambda xf: XUniqWith(pred, xf)



from .private._curry3 import _curry3


def inner_zipWith(fn, a, b):
  rv = []
  idx = 0
  length = min(len(a), len(b))
  while idx < length:
    rv.append(fn(a[idx], b[idx]))
    idx += 1
  return rv


zipWith = _curry3(inner_zipWith)



from .nth import nth
from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._isInteger import _isInteger


def inner_prop(p, obj):
  if obj is None:
    return None
  return nth(p, obj) if _isInteger(p) else getAttribute(obj, p)


prop = _curry2(inner_prop)



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/uniqBy.js
"""


class TestUniqBy(unittest.TestCase):
  def test_returns_a_set_from_any_array_based_on_predicate(self):
    self.assertEqual([-2, -1, 0], R.uniqBy(abs, [-2, -1, 0, 1, 2]))

  def test_keeps_elements_from_the_left(self):
    self.assertEqual([-1, 2, 4, 3], R.uniqBy(abs, [-1, 2, 4, 3, 1, 3]))

  def test_returns_an_empty_array_for_an_empty_array(self):
    self.assertEqual([], R.uniqBy(R.identity, []))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    # self.assertEqual(2, len(R.uniqBy([0], [-0])))
    self.assertEqual(1, len(R.uniqBy(R.identity, [float('nan'), float('nan')])))
    self.assertEqual(1, len(R.uniqBy(R.identity, [Just([1, 2, 3]), Just([1, 2, 3])])))

  def test_can_act_as_a_transducer(self):
    input = [-1, -5, 2, 10, 1, 2]
    expected = [-1, -5, 2, 10]
    self.assertEqual(expected, R.into([], R.uniqBy(abs), input))
    # TODO: R.transduce

if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/uniqWith.js
"""

objs = [
    {'x': R.T, 'i': 0}, {'x': R.F, 'i': 1}, {'x': R.T, 'i': 2}, {'x': R.T, 'i': 3},
    {'x': R.F, 'i': 4}, {'x': R.F, 'i': 5}, {'x': R.T, 'i': 6}, {'x': R.F, 'i': 7},
]
objs2 = [
    {'x': R.T, 'i': 0}, {'x': R.F, 'i': 1}, {'x': R.T, 'i': 2}, {'x': R.T, 'i': 3},
    {'x': R.F, 'i': 0}, {'x': R.T, 'i': 1}, {'x': R.F, 'i': 2}, {'x': R.F, 'i': 3},
]


def eqI(x, accX):
  return x['i'] == accX['i']


class TestUniqWith(unittest.TestCase):
  def test_returns_a_set_from_any_array_based_on_predicate(self):
    self.assertEqual(objs, R.uniqWith(eqI, objs))
    self.assertEqual([{'x': R.T, 'i': 0}, {'x': R.F, 'i': 1}, {'x': R.T, 'i': 2}, {'x': R.T, 'i': 3}], R.uniqWith(eqI, objs2))

  def test_keeps_element_from_the_left(self):
    self.assertEqual([{'i': 1}, {'i': 2}, {'i': 3}, {'i': 4}], R.uniqWith(eqI, [{'i': 1}, {'i': 2}, {'i': 3}, {'i': 4}, {'i': 1}]))

  def test_returns_an_empty_array_for_an_empty_array(self):
    self.assertEqual([], R.uniqWith(eqI, []))

  def test_can_act_as_a_transducer(self):
    input = [1, '1', 2, 1]
    expected = [1, 2]
    # TODO: eqBy
    # TODO: transduce


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/clone.js
"""


class TestCloneIntegersStringAndBooleans(unittest.TestCase):
  def test_clones_integers(self):
    self.assertEqual(-4, R.clone(-4))
    self.assertEqual(9007199254740991, R.clone(9007199254740991))

  def test_clones_floats(self):
    self.assertEqual(-4.5, R.clone(-4.5))
    self.assertEqual(0.0, R.clone(0.0))

  def test_clones_strings(self):
    self.assertEqual('ramda', R.clone('ramda'))

  def test_clones_booleans(self):
    self.assertEqual(True, R.clone(True))


class TestDeepCloneObjects(unittest.TestCase):
  def test_clones_object(self):
    class Obj:
      def __init__(self, x):
        self.value = x
    obj = Obj(42)
    clone = R.clone(obj)
    self.assertEqual(False, obj == clone)
    self.assertEqual(True, isinstance(clone, Obj))

  def test_clones_object_override_eq(self):
    class Obj:
      def __init__(self, x):
        self.value = x

      def __eq__(self, other):
        return self.value == other.value

      def get(self):
        return self.value

      def set(self, value):
        self.value = value

    obj = Obj(10)
    clone = R.clone(obj)
    self.assertEqual(True, obj == clone)

    self.assertEqual(10, obj.get())
    self.assertEqual(10, clone.get())

    obj.set(11)
    self.assertEqual(11, obj.get())
    self.assertEqual(10, clone.get())


class TestDeepCloneDicts(unittest.TestCase):
  def test_clones_shallow_dict(self):
    class Date:
      def __init__(self, year, month, day):
        self.year = year
        self.month = month
        self.day = day

      def setDate(self, day):
        self.day = day

      def __eq__(self, other):
        return self.year == other.year and self.month == other.month and self.day == other.day

    obj = {'a': 1, 'b': 'ramda', 'c': True, 'd': Date(2013, 11, 25)}
    clone = R.clone(obj)
    obj['c'] = False
    obj['d'].setDate(31)
    self.assertEqual({'a': 1, 'b': 'ramda', 'c': True, 'd': Date(2013, 11, 25)}, clone)

  def test_clones_deep_dict(self):
    obj = {'a': {'b': {'c': 'ramda'}}}
    clone = R.clone(obj)
    obj['a']['b']['c'] = None
    self.assertEqual({'a': {'b': {'c': 'ramda'}}}, clone)

  def test_clones_dict_with_circular_references(self):
    x = {'c': None}
    y = {'a': x}
    z = {'b': y}
    x['c'] = z
    clone = R.clone(x)
    # self.assertEqual(x, clone) Python does not support this
    # RecursionError: maximum recursion depth exceeded in comparison


class TestDeepCloneArrays(unittest.TestCase):
  def test_clones_shallow_arrays(self):
    arr = [1, 2, 3]
    clone = R.clone(arr)
    arr.pop()
    self.assertEqual([1, 2, 3], clone)

  def test_clones_deep_arrays(self):
    arr = [1, [1, 2, 3], [[[5]]]]
    clone = R.clone(arr)
    self.assertEqual(arr, clone)
    self.assertEqual(arr[2], clone[2])
    self.assertEqual(arr[2][0], clone[2][0])

    self.assertEqual([1, [1, 2, 3], [[[5]]]], clone)


class TestDeepCloneFunctions(unittest.TestCase):
  def test_keep_reference_to_function(self):
    def fn(x): return x + x
    arr = [{'a': fn}]
    clone = R.clone(arr)
    self.assertEqual(20, clone[0]['a'](10))
    self.assertEqual(clone[0]['a'], arr[0]['a'])

  class TestBuildInTypes(unittest.TestCase):
    # TODO: regex and date is not built-in in Python
    # So we skip the test case for now
    pass


class TestCloneDeepNestedMixedObjects(unittest.TestCase):
  def test_clones_array_with_dicts(self):
    arr = [{'a': {'b': 1}}, [{'c': {'d': 1}}]]
    clone = R.clone(arr)
    arr[1][0] = None
    self.assertEqual([{'a': {'b': 1}}, [{'c': {'d': 1}}]], clone)

  def test_clones_array_with_arrays(self):
    arr = [[1], [[3]]]
    clone = R.clone(arr)
    arr[1][0] = None
    self.assertEqual([[1], [[3]]], clone)

  def test_clones_array_with_mutual_ref_dict(self):
    obj = {'a': 1}
    arr = [{'b': obj}, {'b': obj}]
    clone = R.clone(arr)
    self.assertEqual(arr[0]['b'], clone[1]['b'])
    self.assertEqual(clone[0]['b'], arr[1]['b'])
    self.assertEqual(clone[0]['b'], arr[0]['b'])
    self.assertEqual(clone[1]['b'], arr[1]['b'])

    self.assertEqual({'a': 1}, clone[0]['b'])
    self.assertEqual({'a': 1}, clone[1]['b'])

    obj['a'] = 2
    self.assertEqual({'a': 1}, clone[0]['b'])
    self.assertEqual({'a': 1}, clone[1]['b'])

class TestDeepCloneEdgeCases(unittest.TestCase):
  def test_None_and_empty_dict_and_arrays(self):
    self.assertEqual(None, R.clone(None))
    self.assertEqual({}, R.clone({}))
    self.assertEqual([], R.clone([]))

class TestUseAnArbitraryUserDefinedCloneMethod(unittest.TestCase):
  def test_dispatches_to_clone_method_if_present(self):
    class ArbitraryClone:
      def __init__(self, x):
        self.value = x
      def clone(self):
        return ArbitraryClone(self.value)

    obj = ArbitraryClone(42)
    ArbitraryCloneObj = R.clone(obj)
    self.assertEqual(42, ArbitraryCloneObj.value)
    self.assertEqual(True, isinstance(ArbitraryCloneObj, ArbitraryClone))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/T.js
"""


class TestT(unittest.TestCase):
  def test_always_returns_true(self):
    self.assertEqual(True, R.T())
    self.assertEqual(True, R.T(10))
    self.assertEqual(True, R.T(True))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R
from ramda.private._helper import funcArgsLength

"""
https://github.com/ramda/ramda/blob/master/test/once.js
"""


class TestOnce(unittest.TestCase):
  def test_returns_a_function_that_calls_the_supplied_function_only_the_first_time_called(self):
    ctr = 0
    def wrapper():
      nonlocal ctr
      ctr += 1
    fn = R.once(wrapper)
    fn()
    self.assertEqual(1, ctr)
    fn()
    self.assertEqual(1, ctr)
    fn()
    self.assertEqual(1, ctr)

  def test_passes_along_arguments_supplied(self):
    fn = R.once(lambda a, b: a + b)
    result = fn(5, 10)
    self.assertEqual(15, result)

  def test_retains_and_returns_the_first_value_calculated_even_if_different_arguments_are_passed_later(self):
    ctr = 0

    def add(a, b):
      nonlocal ctr
      ctr += 1
      return a + b
    fn = R.once(add)
    result = fn(5, 10)
    self.assertEqual(15, result)
    self.assertEqual(1, ctr)
    result = fn(20, 30)
    self.assertEqual(15, result)
    self.assertEqual(1, ctr)

  def test_retains_arity(self):
    f = R.once(lambda a, b: a + b)
    self.assertEqual(2, funcArgsLength(f))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xtake import _xtake
from .slice import slice


def inner_take(n, xs):
  return slice(0, None if n < 0 else n, xs)


take = _curry2(_dispatchable(['take'], _xtake, inner_take))




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/match.js
"""

re = r'[A-Z]\d\d\-[a-zA-Z]+'
matching = 'B17-afn'
notMatching = 'B1-afn'


class TestMatch(unittest.TestCase):
  def test_determines_whether_a_string_matches_a_regex(self):
    self.assertEqual(1, len(R.match(re, matching)))
    self.assertEqual(0, len(R.match(re, notMatching)))

  def test_defaults_to_a_different_empty_array_each_time(self):
    first = R.match(re, notMatching)
    second = R.match(re, notMatching)
    self.assertIsNot(first, second)

  def test_throws_on_null_input(self):
    with self.assertRaises(TypeError):
      R.match(re, None)

if __name__ == '__main__':
  unittest.main()



from ._helper import getAttribute
from ._reduced import _reduced
from ._xfBase import XfBase


class XFind(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f
    self.found = False

  def result(self, result):
    if not self.found:
      result = getAttribute(self.xf, '@@transducer/step')(result, None)
    return self.xf.get('@@transducer/result')(result)

  def step(self, result, _input):
    if self.f(_input):
      self.found = True
      result = _reduced(getAttribute(self.xf, '@@transducer/step')(result, _input))
    return result


def _xfind(f): return lambda xf: XFind(f, xf)



from .private._curry2 import _curry2

objOf = _curry2(lambda key, val: {key: val})
"""
Since in JavaScript we can represent key-value as object,
but in Python, we can't.
So we treat objOf only works for dict in Python.
But for the naming convention, we keep the name of objOf.
"""



from typing import Iterable

from ._helper import getAttribute
from ._isArrayLike import _isArrayLike
from ._isFunction import _isFunction


def _createReduce(arrayReduce, methodReduce, iterableReduce):
  def _reduce(xf, acc, arr):
    if arr is None:
      return acc
    if _isArrayLike(arr):
      return arrayReduce(xf, acc, arr)
    if _isFunction(getAttribute(arr, 'fantasy-land/reduce')):
      return methodReduce(xf, acc, arr, 'fantasy-land/reduce')
    if isinstance(arr, Iterable):
      return iterableReduce(xf, acc, arr)
    if _isFunction(getattr(arr, 'reduce', None)):
      return methodReduce(xf, acc, arr, 'reduce')
    raise Exception('reduce: list must be array or iterable')
  return _reduce



import unittest
from datetime import date

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/always.js
"""


class TestAlways(unittest.TestCase):
  def test_returns_a_function_that_returns_the_object_initially_supplied(self):
    theMeaning = R.always(42)
    self.assertEqual(42, theMeaning())
    self.assertEqual(42, theMeaning(10))
    self.assertEqual(42, theMeaning(False))

  def test_works_with_various_types(self):
    self.assertEqual(False, R.always(False)())
    self.assertEqual('abc', R.always('abc')())
    self.assertEqual({'a': 1, 'b': 2}, R.always({'a': 1, 'b': 2})())
    self.assertEqual({'a': 1, 'b': 2}, R.always({'a': 1, 'b': 2})())
    obj = {'a': 1, 'b': 2}
    self.assertEqual(obj, R.always(obj)())
    self.assertEqual(obj, R.always(obj)())
    now = date(1776, 6, 4)
    self.assertEqual(date(1776, 6, 4), R.always(now)())
    self.assertEqual(None, R.always(None)())


class TestAlwaysProperties(unittest.TestCase):
  def test_returns_initial_argument(self):
    a = 42
    b = 27
    f = R.always(a)
    self.assertTrue(f() == a)
    self.assertTrue(f(b) == a)


if __name__ == '__main__':
  unittest.main()



from .private._curry3 import _curry3
from .private._helper import getAttribute


def inner_reduceRight(fn, acc, arr):
  idx = len(arr) - 1
  while idx >= 0:
    acc = fn(arr[idx], acc)
    if acc and getAttribute(acc, '@@transducer/reduced'):
      acc = getAttribute(acc, '@@transducer/value')
      break
    idx -= 1
  return acc


reduceRight = _curry3(inner_reduceRight)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/reduced.js
"""


class TestReduced(unittest.TestCase):
  def test_wraps_a_value(self):
    v = {}
    self.assertEqual(v, R.reduced(v)['@@transducer/value'])

  def test_flags_value_as_reduced(self):
    self.assertEqual(True, R.reduced({})['@@transducer/reduced'])

  def test_short_circuits_reduced(self):
    def fn(acc, v):
      result = acc + v
      if result >= 10:
        result = R.reduced(result)
      return result
    self.assertEqual(10, R.reduce(fn, 0, [1, 2, 3, 4, 5]))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .toString import toString


def inner_max(a, b):
  if a == b:
    return b

  # pylint: disable=unidiomatic-typecheck
  if type(a) != type(b):
    strA = str(a)
    strB = str(b)
    return a if strA > strB else b

  def safeMax(x, y):
    if (x > y) != (y > x):
      return x if x > y else y
    return None
  maxByValue = safeMax(a, b)
  if maxByValue is not None:
    return maxByValue

  stringA = toString(a)
  maxByStringValue = safeMax(stringA, toString(b))
  if maxByStringValue is not None:
    return a if maxByStringValue == stringA else b
  return b


Max = _curry2(inner_max)



from ._helper import getAttribute
from ._xfBase import XfBase


class XMap(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f

  def step(self, result, _input):
    return getAttribute(self.xf, '@@transducer/step')(result, self.f(_input))


def _xmap(f): return lambda xf: XMap(f, xf)



from ._helper import getAttribute
from ._reduced import _reduced
from ._xfBase import XfBase


class XTake(XfBase):
  def __init__(self, n, xf):
    self.xf = xf
    self.n = n
    self.i = 0

  def step(self, result, _input):
    self.i += 1
    ret = result if self.n == 0 else getAttribute(self.xf, '@@transducer/step')(result, _input)
    if self.n >= 0 and self.i >= self.n:
      return _reduced(ret)
    return ret


def _xtake(n): return lambda xf: XTake(n, xf)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/reduceRight.js
"""


def avg(a, b): return (a + b) / 2


def concatFirstThree(val, acc): return R.reduced(acc) if len(acc) == 3 else R.concat(acc, val)


class TestReduceRight(unittest.TestCase):
  def test_folds_lists_in_the_right_order(self):
    self.assertEqual('abcd', R.reduceRight(lambda a, b: a + b, '', ['a', 'b', 'c', 'd']))

  def test_folds_subtract_over_arrays_in_the_right_order(self):
    self.assertEqual(-2, R.reduceRight(lambda a, b: a - b, 0, [1, 2, 3, 4]))

  def test_folds_simple_functions_over_arrays_with_the_supplied_accumulator(self):
    self.assertEqual(12, R.reduceRight(avg, 54, [12, 4, 10, 6]))

  def test_returns_the_accumulator_for_an_empty_array(self):
    self.assertEqual(0, R.reduceRight(avg, 0, []))

  def test_short_circuits_with_reduced(self):
    self.assertEqual('ram', R.reduceRight(concatFirstThree, '', ['a', 'd', 'm', 'a', 'r']))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xall import _xall


def inner_all(fn, arr):
  idx = 0
  while idx < len(arr):
    if not fn(arr[idx]):
      return False
    idx += 1
  return True

# pylint: disable=redefined-builtin
all = _curry2(_dispatchable(['all'], _xall, inner_all))



from .private._clone import _clone
from .private._curryN import _curryN
from .private._dispatchable import _dispatchable
from .private._has import _has
from .private._helper import getAttribute
from .private._reduced import _reduced
from .private._xReduce import _xReduce
from .private._xReduceBy import _xReduceBy
from .private._xwrap import _xwrap


def inner_reduceBy(valueFn, valueAcc, keyFn, arr):
  def inner_xwrap(acc, elt):
    key = keyFn(elt)
    value = valueFn(acc[key] if _has(acc, key) else _clone(valueAcc, deep=False), elt)

    if value and getAttribute(value, '@@transducer/reduced'):
      return _reduced(acc)

    acc[key] = value
    return acc
  xf = _xwrap(inner_xwrap)
  return _xReduce(xf, {}, arr)

reduceBy = _curryN(4, [], _dispatchable([], _xReduceBy, inner_reduceBy))



import copy

from .keys import keys
from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._filter import _filter
from .private._has import _has
from .private._isArrayLike import _isArrayLike
from .private._reduce import _reduce
from .private._xfilter import _xfilter


def inner_filter(pred, filterable):
  if _isArrayLike(filterable):
    return _filter(pred, filterable)
  def inner_reduce(acc, key):
    """
    There are 2 cases of filterable
    case 1: filterable is a dict or an instance with get method
    case 2: filterable is an instance of some classes
    """
    if isinstance(filterable, dict) or _has(filterable, 'get'):
      if pred(filterable.get(key)):
        acc[key] = filterable.get(key)
    else:
      # This is special, because we deepcopy the original object,
      # so we delete attr from original object if not match
      if not pred(getattr(filterable, key, None)):
        delattr(acc, key)
    return acc
  return _reduce(inner_reduce, {} if isinstance(filterable, dict) or _has(filterable, 'get') else copy.deepcopy(filterable), keys(filterable))


# pylint: disable=redefined-builtin
filter = _curry2(_dispatchable(['fantasy-land/filter', 'filter'], _xfilter, inner_filter))




import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/propEq.js
"""

dict1 = {'name': 'Abby', 'age': 7, 'hair': 'blond'}
dict2 = {'name': 'Fred', 'age': 12, 'hair': 'brown'}


class TestPropEq(unittest.TestCase):
  def test_has_R_equals_semantics(self):
    # TODO: handle minus zero
    self.assertEqual(True, R.propEq(float('nan'), 'value', {'value': float('nan')}))
    self.assertEqual(True, R.propEq(Just([42]), 'value', {'value': Just([42])}))

  def test_returns_False_if_called_with_None(self):
    self.assertEqual(False, R.propEq('Abby', 'name', None))

class TestPropEqForDict(unittest.TestCase):
  def test_determines_whether_a_particular_property_matches_a_given_value_for_a_specific_object(self):
    self.assertEqual(True, R.propEq('Abby', 'name', dict1))
    self.assertEqual(True, R.propEq('brown', 'hair', dict2))
    self.assertEqual(False, R.propEq('blond', 'hair', dict2))

class TestPropEqForObject(unittest.TestCase):
  def test_determines_whether_a_particular_property_matches_a_given_value_for_a_specific_object(self):
    class Person:
      def __init__(self, name, age, hair):
        self.name = name
        self.age = age
        self.hair = hair
    abby = Person('Abby', 7, 'blond')
    fred = Person('Fred', 12, 'brown')
    self.assertEqual(True, R.propEq('Abby', 'name', abby))
    self.assertEqual(True, R.propEq('brown', 'hair', fred))
    self.assertEqual(False, R.propEq('blond', 'hair', fred))

class TestPropEqForList(unittest.TestCase):
  def test_handles_number_as_property(self):
    deities=['Cthulhu', 'Dagon', 'Yog-Sothoth']
    self.assertEqual(True, R.propEq('Cthulhu', 0, deities))
    self.assertEqual(True, R.propEq('Dagon', 1, deities))
    self.assertEqual(True, R.propEq('Yog-Sothoth', 2, deities))
    self.assertEqual(True, R.propEq('Yog-Sothoth', -1, deities))
    self.assertEqual(True, R.propEq(None, 3, deities))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/sum.js
"""


class TestSum(unittest.TestCase):
  def test_adds_together_the_array_of_numbers_supplied(self):
    self.assertEqual(10, R.sum([1, 2, 3, 4]))

  def test_does_not_save_the_state_of_the_accumulator(self):
    self.assertEqual(10, R.sum([1, 2, 3, 4]))
    self.assertEqual(1, R.sum([1]))
    self.assertEqual(25, R.sum([5, 5, 5, 5, 5]))

if __name__ == '__main__':
  unittest.main()



from ._helper import getAttribute


class XfBase:
  """
  This is a class which not included in Ramda.
  For extracting the common part to deal with transducer related logic.
  """

  def _init_(self, xf):
    self.xf = xf

  def init(self):
    return getAttribute(self.xf, '@@transducer/init')()

  def result(self, result):
    return getAttribute(self.xf, '@@transducer/result')(result)

  # pylint: disable=no-self-use
  def step(self, result, _input):
    raise Exception('Child class should implement this')

  def get(self, name, default=None):
    if name == '@@transducer/init':
      return self.init
    if name == '@@transducer/result':
      return self.result
    if name == '@@transducer/step':
      return self.step
    return default



import types
import unittest

import ramda as R
from ramda.private._helper import funcArgsLength

"""
https://github.com/ramda/ramda/blob/master/test/compose.js
"""


class TestCompose(unittest.TestCase):
  def test_is_a_variadic_function(self):
    self.assertTrue(type(R.compose) == types.FunctionType)
    self.assertEqual(0, funcArgsLength(R.compose))

  def test_performs_right_to_left_function_composition(self):
    def add(a, b, c): return a + b + c
    def inc(a): return a + 1
    def pow(a): return a ** 2
    f = R.compose(pow, inc, add)
    self.assertEqual(3, funcArgsLength(f))
    self.assertEqual(49, f(1, 2, 3))

  def test_throw_if_given_no_arguments(self):
    with self.assertRaises(Exception):
      R.compose()

  def test_can_be_applied_to_one_argument(self):
    def f(a, b, c): return [a, b, c]
    g = R.compose(f)
    self.assertEqual(3, funcArgsLength(g))
    self.assertEqual([1, 2, 3], g(1, 2, 3))

  def test_composes_two_functions(self):
    def inc(a): return a + 1
    def pow(a): return a ** 2
    self.assertEqual(inc(pow(2)), R.compose(inc, pow)(2))

  def test_associative(self):
    def f(a): return a + 1
    def g(a): return a ** 2
    def h(a): return a * 2
    x = 2
    result = f(g(h(x)))
    self.assertEqual(result, R.compose(f, g, h)(x))
    self.assertEqual(result, R.compose(f, R.compose(g, h))(x))
    self.assertEqual(result, R.compose(R.compose(f, g), h)(x))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xdrop import _xdrop
from .slice import slice


def inner_drop(n, xs):
  return slice(max(0, n), len(xs), xs)


drop = _curry2(_dispatchable(['drop'], _xdrop, inner_drop))



from .curryN import curryN
from .private._curry1 import _curry1
from .private._helper import funcArgsLength


def inner_flip(fn):
  def wrapper(a, b, *args):
    return fn(b, a, *args)
  return curryN(funcArgsLength(fn), wrapper)


flip = _curry1(inner_flip)



import inspect


def _isFunction(fn):
  """
  Python class is also callable, so we need to deal with such pattern.
  class A:
    def b(self):
      return False
  a = A()
  callable(A) # True
  callable(a) # False
  callable(a.b) # True
  inspect.isclass(A) # True
  inspect.isclass(a) # False
  inspect.isclass(a.b) # False
  """
  return callable(fn) and not inspect.isclass(fn)



from ._arity import _arity
from ._isPlaceholder import _isPlaceholder


def _curryN(n, received, fn):
  def f1(*arguments):
    combined = []
    argsIdx = 0
    left = n
    combinedIdx = 0
    while combinedIdx < len(received) or argsIdx < len(arguments):
      result = None
      if combinedIdx < len(received) and ((not _isPlaceholder(received[combinedIdx])) or argsIdx >= len(arguments)):
        result = received[combinedIdx]
      else:
        result = arguments[argsIdx]
        argsIdx += 1
      combined.append(result)
      if not _isPlaceholder(result):
        left -= 1
      combinedIdx += 1
    if left <= 0:
      return fn(*combined)
    return _arity(left, _curryN(n, combined, fn))
  return f1



from ..equals import equals
from ..filter import filter


class _Set(set):
  def add(self, item):
    if filter(equals(item), list(self.__iter__())):
      return False
    super().add(item)
    return True



from ._createReduce import _createReduce
from ._helper import getAttribute
from ._xArrayReduce import _xArrayReduce


def _xIterableReduce(xf, acc, _iter):
  while True:
    try:
      acc = getAttribute(xf, '@@transducer/step')(acc, next(_iter))
      if acc and getAttribute(acc, '@@transducer/reduced'):
        acc = getAttribute(acc, '@@transducer/value')
        break
    except StopIteration:
      break
  return getAttribute(xf, '@@transducer/result')(acc)


def _xMethodReduce(xf, acc, obj, methodName):
  result = getAttribute(xf, '@@transducer/result')
  method = getAttribute(obj, methodName)
  step = getAttribute(xf, '@@transducer/step')
  return result(method(step, acc))


_xReduce = _createReduce(_xArrayReduce, _xMethodReduce, _xIterableReduce)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/xprod.js
"""


class TestXProd(unittest.TestCase):
  def test_returns_an_empty_list_if_either_input_list_is_empty(self):
    self.assertEqual([], R.xprod([], [1, 2, 3]))
    self.assertEqual([], R.xprod([1, 2, 3], []))

  def test_creates_the_collection_of_all_cross_product_pairs_of_its_parameters(self):
    self.assertEqual([[1, 'a'], [1, 'b'], [1, 'c'], [2, 'a'], [2, 'b'], [2, 'c']], R.xprod([1, 2], ['a', 'b', 'c']))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/difference.js
"""

M = [1, 2, 3, 4]
M2 = [1, 2, 3, 4, 1, 2, 3, 4]
N = [3, 4, 5, 6]
N2 = [3, 3, 4, 4, 5, 5, 6, 6]
Z = [3, 4, 5, 6, 10]
Z2 = [1, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 8]


class TestDifference(unittest.TestCase):
  def test_finds_the_set_of_all_elements_in_the_first_list_not_contained_in_the_second(self):
    self.assertEqual([1, 2], R.difference(M, N))

  def test_does_not_allow_duplicates_in_the_output_even_if_the_input_lists_had_duplicates(self):
    self.assertEqual([1, 2], R.difference(M2, N2))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    # self.assertEqual(1, len(R.difference([0], [-0])))
    self.assertEqual(0, len(R.difference([float('nan')], [float('nan')])))
    self.assertEqual(0, len(R.difference([Just([42])], [Just([42])])))

  def test_works_for_array_of_different_lengths(self):
    self.assertEqual([10], R.difference(Z, Z2))
    self.assertEqual([1, 2, 7, 8], R.difference(Z2, Z))

  def test_returns_an_empty_array_if_there_are_no_different_elements(self):
    self.assertEqual([], R.difference(M2, M))
    self.assertEqual([], R.difference(M, M2))
    self.assertEqual([], R.difference([], M2))


if __name__ == '__main__':
  unittest.main()



from ._clone import _clone
from ._has import _has
from ._helper import getAttribute
from ._xfBase import XfBase


class XReduceBy(XfBase):
  def __init__(self, valueFn, valueAcc, keyFn, xf):
    self.valueFn = valueFn
    self.valueAcc = valueAcc
    self.keyFn = keyFn
    self.xf = xf
    self.inputs = {}

  def result(self, result):
    for key in self.inputs:
      if _has(self.inputs, key):
        result = getAttribute(self.xf, '@@transducer/step')(result, self.inputs[key])
        if getAttribute(result, '@@transducer/reduced'):
          result = getAttribute(result, '@@transducer/value')
          break
    self.inputs = None
    return getAttribute(self.xf, '@@transducer/result')(result)

  def step(self, result, _input):
    key = self.keyFn(_input)
    self.inputs[key] = getAttribute(self.inputs, key) or [key, _clone(self.valueAcc, deep=False)]
    self.inputs[key][1] = self.valueFn(self.inputs[key][1], _input)
    return result


def _xReduceBy(valueFn, valueAcc, keyFn): return lambda xf: XReduceBy(valueFn, valueAcc, keyFn, xf)



from .curryN import curryN
from .private._curry2 import _curry2


def inner_useWith(fn, transformers):
  def wrapper(*arguments):
    args = []
    idx = 0
    while idx < len(transformers):
      args.append(transformers[idx](arguments[idx]))
      idx += 1
    return fn(*(args + list(arguments[len(transformers):])))
  return curryN(len(transformers), wrapper)


useWith = _curry2(inner_useWith)



from ._helper import getAttribute
from ._reduced import _reduced
from ._xfBase import XfBase


class XAny(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f
    self.any = False

  def result(self, result):
    if not self.any:
      result = getAttribute(self.xf, '@@transducer/step')(result, False)
    return self.xf.get('@@transducer/result')(result)

  def step(self, result, _input):
    if self.f(_input):
      self.any = True
      result = _reduced(getAttribute(self.xf, '@@transducer/step')(result, True))
    return result


def _xany(f): return lambda xf: XAny(f, xf)



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._includesWith import _includesWith
from .private._xuniqWith import _xuniqWith


def inner_uniqWith(pred, arr):
  idx = 0
  length = len(arr)
  result = []
  while idx < length:
    item = arr[idx]
    if not _includesWith(pred, item, result):
      result.append(item)
    idx += 1
  return result


uniqWith = _curry2(_dispatchable([], _xuniqWith, inner_uniqWith))



import unittest
from functools import cmp_to_key

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/comparator.js
"""


class TestComparator(unittest.TestCase):
  def test_builds_a_comparator_function_for_sorting_out_of_a_simple_predicate_that_reports_whether_the_first_param_is_smaller(self):
    self.assertEqual([1, 1, 2, 3, 5, 8], sorted([3, 1, 8, 1, 2, 5], key=cmp_to_key(R.comparator(lambda a, b: a < b))))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._isNumber import _isNumber


def inner_range(frm, to):
  if not (_isNumber(frm) and _isNumber(to)):
    raise TypeError('Both arguments to range must be numbers')
  result = []
  n = frm
  while n < to:
    result.append(n)
    n += 1
  return result


# pylint: disable=redefined-builtin
range = _curry2(inner_range)



from ..__ import __
from ._isPlaceholder import _isPlaceholder


def _curry1(fn):
  # pylint: disable=dangerous-default-value
  # pylint: disable=keyword-arg-before-vararg
  def f1(a=__, *_):
    if _isPlaceholder(a):
      return f1
    return fn(a)
  return f1



from ._curry1 import _curry1
from ._isPlaceholder import _isPlaceholder


def _curry2(fn):
  def f2(*args):
    def f_b(_b): return fn(args[0], _b)
    if len(args) == 0:
      return f2
    if len(args) == 1:
      a = args[0]
      if _isPlaceholder(a):
        return f2

      return _curry1(f_b)
    a, b = args[0], args[1]
    if _isPlaceholder(a) and _isPlaceholder(b):
      return f2
    if _isPlaceholder(a):
      def f_a(_a): return fn(_a, b)
      return _curry1(f_a)
    if _isPlaceholder(b):
      return _curry1(f_b)
    return fn(a, b)
  return f2



from math import isnan

from ..equals import equals
from ._helper import getAttribute
from ._isFunction import _isFunction
from ._isNumber import _isNumber


def _indexOf(arr, a, idx):
  if _isFunction(getAttribute(arr, 'indexOf')):
    if _isNumber(a) and isnan(a):
      while idx < len(arr):
        if isnan(arr[idx]):
          return idx
        idx += 1
    return arr.indexOf(a, idx)

  while idx < len(arr):
    if equals(arr[idx], a):
      return idx
    idx += 1
  return -1



from ._helper import getAttribute
from ._xfBase import XfBase


class XFilter(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f

  def step(self, result, _input):
    return getAttribute(self.xf, '@@transducer/step')(result, _input) if self.f(_input) else result

def _xfilter(f): return lambda xf: XFilter(f, xf)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/path.js
"""

deepDict = {
    'a': {
        'b': {
            'c': 'c',
        }
    },
    'falseVal': False,
    'nullVal': None,
    'arrayVal': ['arr'],
}


class TestPathForDict(unittest.TestCase):
  def test_takes_a_path_and_an_object_and_returns_the_value_at_the_path_or_None(self):
    obj = {
        'a': {
            'b': {
                'c': 100,
                'd': 200,
            },
            'e': {
                'f': [100, 101, 102],
                'g': 'G',
            },
            'h': 'H',
        },
        'i': 'I',
        'j': ['J']
    }
    self.assertEqual(100, R.path(['a', 'b', 'c'], obj))
    self.assertEqual(obj, R.path([], obj))
    self.assertEqual(101, R.path(['a', 'e', 'f', 1], obj))
    self.assertEqual('J', R.path(['j', 0], obj))
    self.assertEqual(None, R.path(['j', 1], obj))

  def test_takes_a_path_that_contains_indices_into_arrays(self):
    obj = {
        'a': [[{}], [{'x': 'first'}, {'x': 'second'}, {'x': 'third'}, {'x': 'last'}]]
    }
    self.assertEqual({}, R.path(['a', 0, 0], obj))
    self.assertEqual(None, R.path(['a', 0, 1], obj))
    self.assertEqual('first', R.path(['a', 1, 0, 'x'], obj))
    self.assertEqual('second', R.path(['a', 1, 1, 'x'], obj))
    self.assertEqual('A', R.path([0], ['A']))

  def test_takes_a_path_that_contains_negative_indices_into_arrays(self):
    self.assertEqual('c', R.path(['x', -2], {'x': ['a', 'b', 'c', 'd']}))
    self.assertEqual(97, R.path([-1, 'y'], [{'x': 1, 'y': 99}, {'x': 2, 'y': 98}, {'x': 3, 'y': 97}]))

  def test_gets_a_deep_property_value_from_dicts(self):
    self.assertEqual('c', R.path(['a', 'b', 'c'], deepDict))
    self.assertEqual(deepDict['a'], R.path(['a'], deepDict))

  def test_returns_None_for_items_not_found(self):
    self.assertEqual(None, R.path(['a', 'b', 'foo'], deepDict))
    self.assertEqual(None, R.path(['bar'], deepDict))
    self.assertEqual(None, R.path(['a', 'b'], {'a': None}))


class Obj:
  def __init__(self, a, falseVal, nullVal, arrayVal):
    self.a = a
    self.falseVal = falseVal
    self.nullVal = nullVal
    self.arrayVal = arrayVal


class A:
  def __init__(self, b):
    self.b = b


class B:
  def __init__(self, c):
    self.c = c


b = B('c')
a = A(b)
deepObject = Obj(a, False, None, ['arr'])


class TestPathForObject(unittest.TestCase):
  def test_takes_a_path_and_an_object_and_returns_the_value_at_the_path_or_None(self):
    class Obj:
      def __init__(self, a, i, j):
        self.a = a
        self.i = i
        self.j = j

    class A:
      def __init__(self, b, e, h):
        self.b = b
        self.e = e
        self.h = h

    class B:
      def __init__(self, c, d):
        self.c = c
        self.d = d

    class E:
      def __init__(self, f, g):
        self.f = f
        self.g = g
    b = B(c=100, d=200)
    e = E(f=[100, 101, 102], g='G')
    a = A(b=b, e=e, h='H')
    obj = Obj(a=a, i='I', j=['J'])
    self.assertEqual(100, R.path(['a', 'b', 'c'], obj))
    self.assertEqual(obj, R.path([], obj))
    self.assertEqual(101, R.path(['a', 'e', 'f', 1], obj))
    self.assertEqual('J', R.path(['j', 0], obj))
    self.assertEqual(None, R.path(['j', 1], obj))

  def test_takes_a_path_that_contains_indices_into_arrays(self):
    class Obj:
      def __init__(self, a):
        self.a = a

    class A:
      def __init__(self, x):
        self.x = x
    a = [[{}], [A('first'), A('second'), A('third'), A('last')]]
    obj = Obj(a=a)
    self.assertEqual({}, R.path(['a', 0, 0], obj))
    self.assertEqual(None, R.path(['a', 0, 1], obj))
    self.assertEqual('first', R.path(['a', 1, 0, 'x'], obj))
    self.assertEqual('second', R.path(['a', 1, 1, 'x'], obj))
    self.assertEqual('A', R.path([0], ['A']))

  def test_takes_a_path_that_contains_negative_indices_into_arrays(self):
    class Obj:
      def __init__(self, x):
        self.x = x

    class XY:
      def __init__(self, x, y):
        self.x = x
        self.y = y
    self.assertEqual('c', R.path(['x', -2], Obj(['a', 'b', 'c', 'd'])))
    self.assertEqual(97, R.path([-1, 'y'], [XY(x=1, y=99), XY(x=2, y=98), XY(x=3, y=97)]))

  def test_gets_a_deep_property_value_from_dicts(self):
    self.assertEqual('c', R.path(['a', 'b', 'c'], deepObject))
    self.assertEqual(deepObject.a, R.path(['a'], deepObject))

  def test_returns_None_for_items_not_found(self):
    class A:
      def __init__(self, v):
        self.v = v
    self.assertEqual(None, R.path(['a', 'b', 'foo'], deepObject))
    self.assertEqual(None, R.path(['bar'], deepObject))
    self.assertEqual(None, R.path(['a', 'b'], A(None)))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/head.js
"""


class TestHead(unittest.TestCase):
  def test_returns_the_first_element_of_an_ordered_collection(self):
    self.assertEqual(1, R.head([1, 2, 3]))
    self.assertEqual(2, R.head([2, 3]))
    self.assertEqual(3, R.head([3]))
    self.assertEqual(None, R.head([]))

    self.assertEqual('a', R.head('abc'))
    self.assertEqual('b', R.head('bc'))
    self.assertEqual('c', R.head('c'))
    self.assertEqual('', R.head(''))

  def test_throws_if_applied_to_None(self):
    with self.assertRaises(TypeError):
      R.head(None)

if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/eqProps.js
"""

a = []
b = a


class TestEqProps(unittest.TestCase):
  def test_reports_whether_two_dicts_have_the_same_value_for_a_given_property(self):
    self.assertEqual(True, R.eqProps('name', {'name': 'fred', 'age': 10}, {'name': 'fred', 'age': 12}))
    self.assertEqual(False, R.eqProps('name', {'name': 'fred', 'age': 10}, {'name': 'franny', 'age': 10}))

  def test_reports_whether_two_objects_have_the_same_value_for_a_given_property(self):
    class Person:
      def __init__(self, name, age):
        self.name = name
        self.age = age
    fred1 = Person('fred', 10)
    fred2 = Person('fred', 12)
    franny = Person('franny', 10)
    self.assertEqual(True, R.eqProps('name', fred1, fred2))
    self.assertEqual(False, R.eqProps('name', fred1, franny))

  def test_has_R_equals_semantics(self):
    self.assertEqual(True, R.eqProps('value', {'value': Just([42])}, {'value': Just([42])}))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/paths.js
"""

d = {
    'a': {

        'b': {
            'c': 1,
            'd': 2,
        }
    },
    'p': [{'q': 3}, 'Hi'],
    'x': {
        'y': 'Alice',
        'z': [[{}]]
    }
}


class TestPathsForDict(unittest.TestCase):
  def test_takes_paths_and_returns_values_at_those_paths(self):
    self.assertEqual([1, 'Alice'], R.paths([['a', 'b', 'c'], ['x', 'y']], d))
    self.assertEqual([2, None], R.paths([['a', 'b', 'd'], ['p', 'q']], d))

  def test_takes_a_paths_that_contains_indices_into_arrays(self):
    self.assertEqual([3, {}], R.paths([['p', 0, 'q'], ['x', 'z', 0, 0]], d))
    self.assertEqual([3, None], R.paths([['p', 0, 'q'], ['x', 'z', 2, 1]], d))

  def test_takes_a_paths_that_contains_negative_indices_into_array(self):
    self.assertEqual([3, 'Hi'], R.paths([['p', -2, 'q'], ['p', -1]], d))
    self.assertEqual([None, {}], R.paths([['p', -4, 'q'], ['x', 'z', -1, 0]], d))

  def test_gets_a_deep_property_value_from_dicts(self):
    self.assertEqual([d['a']['b']], R.paths([['a', 'b']], d))
    self.assertEqual([d['p'][0]], R.paths([['p', 0]], d))

  def test_returns_None_for_items_not_found(self):
    self.assertEqual([None], R.paths([['a', 'x', 'y']], d))
    self.assertEqual([None], R.paths([['p', 2]], d))


class Obj:
  def __init__(self, a, p, x):
    self.a = a
    self.p = p
    self.x = x


class A:
  def __init__(self, b):
    self.b = b


class B:
  def __init__(self, c, d):
    self.c = c
    self.d = d


class Q:
  def __init__(self, q):
    self.q = q


class X:
  def __init__(self, y, z):
    self.y = y
    self.z = z


b = B(c=1, d=2)
a = A(b=b)
q = Q(q=3)
p = [q, 'Hi']
x = X(y='Alice', z=[[{}]])
obj = Obj(a=a, p=p, x=x)


class TestPathsForObject(unittest.TestCase):
  def test_takes_paths_and_returns_values_at_those_paths(self):
    self.assertEqual([1, 'Alice'], R.paths([['a', 'b', 'c'], ['x', 'y']], obj))
    self.assertEqual([2, None], R.paths([['a', 'b', 'd'], ['p', 'q']], obj))

  def test_takes_a_paths_that_contains_indices_into_arrays(self):
    self.assertEqual([3, {}], R.paths([['p', 0, 'q'], ['x', 'z', 0, 0]], obj))
    self.assertEqual([3, None], R.paths([['p', 0, 'q'], ['x', 'z', 2, 1]], obj))

  def test_takes_a_paths_that_contains_negative_indices_into_array(self):
    self.assertEqual([3, 'Hi'], R.paths([['p', -2, 'q'], ['p', -1]], obj))
    self.assertEqual([None, {}], R.paths([['p', -4, 'q'], ['x', 'z', -1, 0]], obj))

  def test_gets_a_deep_property_value_from_objects(self):
    self.assertEqual([obj.a.b], R.paths([['a', 'b']], obj))
    self.assertEqual([obj.p[0]], R.paths([['p', 0]], obj))

  def test_returns_None_for_items_not_found(self):
    self.assertEqual([None], R.paths([['a', 'x', 'y']], obj))
    self.assertEqual([None], R.paths([['p', 2]], obj))


if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xtakeWhile import _xtakeWhile
from .slice import slice


def inner_takeWhile(fn, xs):
  idx = 0
  length = len(xs)
  while idx < length and fn(xs[idx]):
    idx += 1
  return slice(0, idx, xs)


takeWhile = _curry2(_dispatchable(['takeWhile'], _xtakeWhile, inner_takeWhile))



from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._indexOf import _indexOf
from .private._isArray import _isArray
from .private._isFunction import _isFunction


def inner_indexOf(target, xs):
  if _isFunction(getAttribute(xs, 'indexOf')) and not _isArray(xs):
    return xs.indexOf(target)
  return _indexOf(xs, target, 0)


indexOf = _curry2(inner_indexOf)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/project.js
"""

kids = [
    {'name': 'Abby', 'age': 7, 'hair': 'blond'},
    {'name': 'Fred', 'age': 12, 'hair': 'brown'},
    {'name': 'Rusty', 'age': 10, 'hair': 'brown'},
    {'name': 'Alois', 'age': 15, 'disposition': 'surly'},
]


class TestProject(unittest.TestCase):
  def test_selects_the_chosen_properties_from_each_element_in_a_list(self):
    self.assertEqual([
        {'name': 'Abby', 'age': 7},
        {'name': 'Fred', 'age': 12},
        {'name': 'Rusty', 'age': 10},
        {'name': 'Alois', 'age': 15},
    ], R.project(['name', 'age'], kids))

  def test_has_a_None_property_on_the_output_tuple_for_any_input_tuple_that_does_not_have_the_property(self):
    self.assertEqual([
        {'name': 'Abby', 'hair': 'blond'},
        {'name': 'Fred', 'hair': 'brown'},
        {'name': 'Rusty', 'hair': 'brown'},
        {'name': 'Alois', 'hair': None},
    ], R.project(['name', 'hair'], kids))

  def test_works_with_objects(self):
    class Human:
      def getName(self):
        return self.name

    class Kid(Human):
      def __init__(self, name, age, hair):
        self.name = name
        self.age = age
        self.hair = hair

      def fn(self):
        return self.name

    class Kid2(Human):
      def __init__(self, name, age, disposition):
        self.name = name
        self.age = age
        self.disposition = disposition

      def fn(self):
        return self.name

    abby = Kid('Abby', 7, 'blond')
    fred = Kid('Fred', 12, 'brown')
    rusty = Kid('Rusty', 10, 'brown')
    alois = Kid2('Alois', 15, 'surly')
    actual = R.project(['name', 'hair'], [abby, fred, rusty, alois])

    self.assertEqual('Abby', actual[0].name)
    self.assertEqual('blond', actual[0].hair)
    self.assertEqual('Fred', actual[1].name)
    self.assertEqual('brown', actual[1].hair)
    self.assertEqual('Rusty', actual[2].name)
    self.assertEqual('brown', actual[2].hair)
    self.assertEqual('Alois', actual[3].name)
    self.assertEqual(None, actual[3].hair)

    self.assertEqual('Abby', actual[0].fn()) # remain methods
    self.assertEqual('Alois', actual[3].fn()) # remain methods
    self.assertEqual('Abby', actual[0].getName()) # remain inherited methods
    self.assertEqual('Abby', actual[0].getName()) # remain inherited methods

if __name__ == '__main__':
  unittest.main()



from ._helper import getAttribute
from ._isArray import _isArray
from ._isFunction import _isFunction
from ._isTransformer import _isTransformer


def _dispatchable(methodNames, transducerCreator, fn):
  def f(*arguments):
    if len(arguments) == 0:
      return fn()
    obj = arguments[-1]
    if not _isArray(obj):
      idx = 0
      while idx < len(methodNames):
        method = getAttribute(obj, methodNames[idx])
        if _isFunction(method):
          return method(*arguments[:-1])
        idx += 1
      if _isTransformer(obj):
        transducer = transducerCreator(*arguments[:-1])
        return transducer(obj)
    return fn(*arguments)
  return f




import unittest

import ramda as R
from ramda.private._curry3 import _curry3

"""
https://github.com/ramda/ramda/blob/master/test/internal/_curry3.js
"""


class Test_Curry3(unittest.TestCase):
  def test_supports_placeholder(self):
    def f(a, b, c): return [a, b, c]
    g = _curry3(f)
    _ = R.__

    self.assertEqual([1, 2, 3], g(1)(2)(3))
    self.assertEqual([1, 2, 3], g(1)(2, 3))
    self.assertEqual([1, 2, 3], g(1, 2)(3))
    self.assertEqual([1, 2, 3], g(1, 2, 3))

    self.assertEqual([1, 2, 3], g(_, 2, 3)(1))
    self.assertEqual([1, 2, 3], g(1, _, 3)(2))
    self.assertEqual([1, 2, 3], g(1, 2, _)(3))

    self.assertEqual([1, 2, 3], g(1, _, _)(2)(3))
    self.assertEqual([1, 2, 3], g(_, 2, _)(1)(3))
    self.assertEqual([1, 2, 3], g(_, _, 3)(1)(2))

    self.assertEqual([1, 2, 3], g(1, _, _)(2, 3))
    self.assertEqual([1, 2, 3], g(_, 2, _)(1, 3))
    self.assertEqual([1, 2, 3], g(_, _, 3)(1, 2))

    self.assertEqual([1, 2, 3], g(1, _, _)(_, 3)(2))
    self.assertEqual([1, 2, 3], g(_, 2, _)(_, 3)(1))
    self.assertEqual([1, 2, 3], g(_, _, 3)(_, 2)(1))

    self.assertEqual([1, 2, 3], g(_, _, _)(_, _)(1, 2, 3))
    self.assertEqual([1, 2, 3], g(_, _, _)(1, _, _)(_, _)(2, _)(_)(3))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/multiply.js
"""


class TestMultiply(unittest.TestCase):
  def test_multiplies_together_two_numbers(self):
    self.assertEqual(42, R.multiply(6, 7))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/sort.js
"""


class TestSort(unittest.TestCase):
  def test_sorts_the_elements_of_a_list(self):
    self.assertEqual([1, 1, 2, 3, 5, 8], R.sort(lambda a, b: a - b, [3, 1, 8, 1, 2, 5]))

  def test_does_not_affect_the_list_passed_supplied(self):
    arr = [3, 1, 8, 1, 2, 5]
    self.assertEqual([1, 1, 2, 3, 5, 8], R.sort(lambda a, b: a - b, arr))
    self.assertEqual([3, 1, 8, 1, 2, 5], arr)


if __name__ == '__main__':
  unittest.main()



from .equals import equals
from .private._curry3 import _curry3
from .private._helper import getAttribute


def inner_eqProps(prop, obj1, obj2):
  return equals(getAttribute(obj1, prop), getAttribute(obj2, prop))


eqProps = _curry3(inner_eqProps)




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/join.js
"""


class TestJoin(unittest.TestCase):
  def test_concatenates_a_list_elements_to_a_string_with_an_separator_string_between_elements(self):
    arr = [1,2,3,4]
    self.assertEqual('1~2~3~4', R.join('~', arr))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/not.js
"""


class TestNot(unittest.TestCase):
  def test_reverses_argument(self):
    self.assertEqual(True, R.Not(False))
    self.assertEqual(False, R.Not(1))
    self.assertEqual(True, R.Not(''))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R
from ramda.private._curry2 import _curry2

"""
https://github.com/ramda/ramda/blob/master/test/internal/_curry2.js
"""


class Test_Curry2(unittest.TestCase):
  def test_supports_placeholder(self):
    def f(a, b): return [a, b]
    g = _curry2(f)
    _ = R.__

    self.assertEqual([1, 2], g(1)(2))
    self.assertEqual([1, 2], g(1, 2))

    self.assertEqual([1, 2], g(_, 2)(1))
    self.assertEqual([1, 2], g(1, _)(2))

    self.assertEqual([1, 2], g(_, _)(1)(2))
    self.assertEqual([1, 2], g(_, _)(1, 2))
    self.assertEqual([1, 2], g(_, _)(_)(1, 2))
    self.assertEqual([1, 2], g(_, _)(_, 2)(1))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/F.js
"""


class TestF(unittest.TestCase):
  def test_always_returns_false(self):
    self.assertEqual(False, R.F())
    self.assertEqual(False, R.F(10))
    self.assertEqual(False, R.F(True))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R
from ramda.private._helper import funcArgsLength
from ramda.private._isFunction import _isFunction

"""
https://github.com/ramda/ramda/blob/master/test/useWith.js
"""

add1 = R.add(1)
mult2 = R.multiply(2)
div3 = R.divide(R.__, 3)
f = R.useWith(max, [add1, mult2, div3])


class TestUseWith(unittest.TestCase):
  def test_takes_a_list_of_function_and_returns_a_function(self):
    self.assertTrue(_isFunction(R.useWith(max, [])))
    self.assertTrue(_isFunction(R.useWith(max, [add1])))
    self.assertTrue(_isFunction(R.useWith(max, [add1, mult2, div3])))

  def test_passes_the_arguments_received_to_their_respective_functions(self):
    self.assertEqual(16, f(7, 8, 9))

  def test_passes_additional_arguments_to_the_main_function(self):
    self.assertEqual(16, f(7, 8, 9, 10))
    self.assertEqual(20, f(7, 8, 9, 20))

  def test_has_the_correct_arity(self):
    self.assertEqual(3, funcArgsLength(f))

if __name__ == '__main__':
  unittest.main()



from ._has import _has
from ._isFunction import _isFunction


def _isTransformer(v):
  """
  We treat transformer as a dict in Python
  """
  return (v is not None) and (_has(v, 'get')) and (_isFunction(v.get('@@transducer/step')))



from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._xfind import _xfind


def inner_find(fn, arr):
  idx = 0
  while idx < len(arr):
    if fn(arr[idx]):
      return arr[idx]
    idx += 1


find = _curry2(_dispatchable(['find'], _xfind, inner_find))



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/groupBy.js
"""

arr = ['foo', 'bar', 'baz', 'quux']


class TestGroupBy(unittest.TestCase):
  def test_splits_the_list_into_groups_according_to_the_grouping_function(self):
    def grade(score):
      if score < 65:
        return 'F'
      if score < 70:
        return 'D'
      if score < 80:
        return 'C'
      if score < 90:
        return 'B'
      return 'A'
    students = [
        {'name': 'Abby', 'score': 84},
        {'name': 'Brad', 'score': 73},
        {'name': 'Chris', 'score': 89},
        {'name': 'Dianne', 'score': 99},
        {'name': 'Eddy', 'score': 58},
        {'name': 'Fred', 'score': 67},
        {'name': 'Gillian', 'score': 91},
        {'name': 'Hannah', 'score': 78},
        {'name': 'Irene', 'score': 85},
        {'name': 'Jack', 'score': 69}
    ]

    def byGrade(student):
      return grade(student.get('score', 0))

    self.assertEqual({
        'A': [{'name': 'Dianne', 'score': 99}, {'name': 'Gillian', 'score': 91}],
        'B': [{'name': 'Abby', 'score': 84}, {'name': 'Chris', 'score': 89}, {'name': 'Irene', 'score': 85}],
        'C': [{'name': 'Brad', 'score': 73}, {'name': 'Hannah', 'score': 78}],
        'D': [{'name': 'Fred', 'score': 67}, {'name': 'Jack', 'score': 69}],
        'F': [{'name': 'Eddy', 'score': 58}]
    }, R.groupBy(byGrade, students))

  def test_returns_an_empty_object_if_given_an_empty_list(self):
    self.assertEqual({}, R.groupBy(R.prop('x'), []))

  def test_dispatches_on_transformer_objects_in_list_position(self):
    pass
    # TODO: R.mergeRight

  def test_can_act_as_a_transducer(self):
    def evenOdd(x):
      return 'even' if x % 2 == 0 else 'odd'
    expected = {'even': [2, 4, 6, 8], 'odd': [1, 3, 5, 7, 9]}
    self.assertEqual(expected, R.into({}, R.groupBy(evenOdd), R.range(1, 10)))
    # TODO: R.transduce


if __name__ == '__main__':
  unittest.main()



def transducer_init():
  raise Exception('init not implemented on XWrap')


def _xwrap(fn):
  return {
      '@@transducer/init': transducer_init,
      '@@transducer/result': lambda acc: acc,
      '@@transducer/step': fn
  }



from ._helper import getAttribute
from ._Set import _Set
from ._xfBase import XfBase


class XUniqBy(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f
    self._set = _Set()

  def step(self, result, _input):
    if self._set.add(self.f(_input)):
      return getAttribute(self.xf, '@@transducer/step')(result, _input)
    return result


def _xuniqBy(f): return lambda xf: XUniqBy(f, xf)



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/tail.js
"""


class TestTail(unittest.TestCase):
  def test_returns_the_tail_of_an_ordered_collection(self):
    self.assertEqual([2, 3], R.tail([1, 2, 3]))
    self.assertEqual([3], R.tail([2, 3]))
    self.assertEqual([], R.tail([3]))
    self.assertEqual([], R.tail([]))

    self.assertEqual('bc', R.tail('abc'))
    self.assertEqual('c', R.tail('bc'))
    self.assertEqual('', R.tail('c'))
    self.assertEqual('', R.tail(''))

  def test_exception_if_work_with_none(self):
    with self.assertRaises(TypeError):
      R.tail(None)


if __name__ == '__main__':
  unittest.main()



from .equals import equals
from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._isArray import _isArray
from .private._isFunction import _isFunction


def inner_lastIndexOf(target, xs):
  if _isFunction(getAttribute(xs, 'lastIndexOf')) and not _isArray(xs):
    return xs.lastIndexOf(target)
  idx = len(xs) - 1
  while idx >= 0:
    if equals(xs[idx], target):
      return idx
    idx -= 1
  return -1


lastIndexOf = _curry2(inner_lastIndexOf)




import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/union.js
"""

M = [1, 2, 3, 4]
N = [3, 4, 5, 6]


class TestUnion(unittest.TestCase):
  def test_combines_two_lists_into_the_set_of_all_their_elements(self):
    self.assertEqual([1, 2, 3, 4, 5, 6], R.union(M, N))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    self.assertEqual(1, len(R.union([float('nan')], [float('nan')])))
    self.assertEqual(1, len(R.union([Just([42])], [Just([42])])))


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/lte.js
"""


class TestLte(unittest.TestCase):
  def test_reports_whether_one_item_is_less_than_or_equal_to_another(self):
    self.assertEqual(True, R.lte(3, 5))
    self.assertEqual(False, R.lte(6, 5))
    self.assertEqual(True, R.lte(7.0, 7.0))
    self.assertEqual(True, R.lte('abc', 'xyz'))
    self.assertEqual(False, R.lte('abcd', 'abc'))


if __name__ == '__main__':
  unittest.main()



from .map import map
from .nth import nth
from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._isInteger import _isInteger


def inner_paths(pathsArray, obj):
  def mapWrapper(_paths):
    val = obj
    idx = 0
    while idx < len(_paths):
      if val is None:
        return None
      p = _paths[idx]
      val = nth(p, val) if _isInteger(p) else getAttribute(val, p)
      idx += 1
    return val
  return map(mapWrapper, pathsArray)


paths = _curry2(inner_paths)



from .private._curry2 import _curry2


def inner_xprod(a, b):
  idx = 0
  ilen = len(a)
  jlen = len(b)
  result = []
  while idx < ilen:
    j = 0
    while j < jlen:
      result.append([a[idx], b[j]])
      j += 1
    idx += 1
  return result


xprod = _curry2(inner_xprod)



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/range.js
"""


class TestRange(unittest.TestCase):
  def test_returns_list_of_numbers(self):
    self.assertEqual([0, 1, 2, 3, 4], R.range(0, 5))
    self.assertEqual([4, 5, 6], R.range(4, 7))

  def test_returns_the_empty_list_if_the_first_parameter_is_not_larger_than_the_second(self):
    self.assertEqual([], R.range(7, 3))
    self.assertEqual([], R.range(5, 5))

  def test_returns_an_empty_array_if_from_larger_than_to(self):
    result = R.range(10, 5)
    self.assertEqual([], result)
    result.append(5)
    self.assertEqual([], R.range(10, 5))

  def test_terminates_given_bad_input(self):
    with self.assertRaises(TypeError):
      R.range('a', 'z')


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/concat.js
"""


class Z1:
  x = 'z1'

  def concat(self, that):
    return self.x + ' ' + that.x


class Z2:
  x = 'z2'


z1 = Z1()
z2 = Z2()


class TestConcat(unittest.TestCase):
  def test_adds_combines_the_elements_of_two_lists(self):
    self.assertEqual(['a', 'b', 'c', 'd'], R.concat(['a', 'b'], ['c', 'd']))
    self.assertEqual(['c', 'd'], R.concat([], ['c', 'd']))

  def test_works_with_strings(self):
    self.assertEqual('foobar', R.concat('foo', 'bar'))
    self.assertEqual('x', R.concat('', 'x'))
    self.assertEqual('x', R.concat('x', ''))
    self.assertEqual('', R.concat('', ''))

  def test_delegates_to_non_strings_object_with_a_concat_method_as_second_param(self):
    self.assertEqual('z1 z2', R.concat(z1, z2))

  def test_throws_if_attemping_to_combine_an_array_with_a_non_array(self):
    with self.assertRaises(Exception):
      R.concat([1], 2)

  def test_throws_if_not_an_array_or_object_with_a_concat_method(self):
    with self.assertRaises(Exception):
      R.concat(z2, z1)

if __name__ == '__main__':
  unittest.main()



from .__ import __
from .add import add
from .adjust import adjust
from .all import all
from .always import always
from .And import And
from .any import any
from .append import append
from .clone import clone
from .comparator import comparator
from .compose import compose
from .concat import concat
from .countBy import countBy
from .curry import curry
from .curryN import curryN
from .difference import difference
from .differenceWith import differenceWith
from .divide import divide
from .drop import drop
from .empty import empty
from .eqProps import eqProps
from .equals import equals
from .F import F
from .filter import filter
from .find import find
from .flatten import flatten
from .flip import flip
from .groupBy import groupBy
from .gt import gt
from .gte import gte
from .head import head
from .identity import identity
from .indexOf import indexOf
from .intersection import intersection
from .into import into
from .invoker import invoker
from .isEmpty import isEmpty
from .join import join
from .keys import keys
from .lastIndexOf import lastIndexOf
from .lt import lt
from .lte import lte
from .map import map
from .match import match
from .Max import Max
from .Min import Min
from .multiply import multiply
from .Not import Not
from .nth import nth
from .objOf import objOf
from .omit import omit
from .once import once
from .Or import Or
from .path import path
from .paths import paths
from .pick import pick
from .pickAll import pickAll
from .pipe import pipe
from .pluck import pluck
from .prepend import prepend
from .product import product
from .project import project
from .prop import prop
from .propEq import propEq
from .props import props
from .range import range
from .reduce import reduce
from .reduceBy import reduceBy
from .reduced import reduced
from .reduceRight import reduceRight
from .reject import reject
from .reverse import reverse
from .slice import slice
from .sort import sort
from .sortBy import sortBy
from .split import split
from .subtract import subtract
from .sum import sum
from .T import T
from .tail import tail
from .take import take
from .takeWhile import takeWhile
from .tap import tap
from .toString import toString
from .union import union
from .unionWith import unionWith
from .uniq import uniq
from .uniqBy import uniqBy
from .uniqWith import uniqWith
from .useWith import useWith
from .values import values
from .xprod import xprod
# pylint: disable=redefined-builtin
from .zip import zip
from .zipWith import zipWith




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/lt.js
"""


class TestLt(unittest.TestCase):
  def test_reports_whether_one_item_is_less_than_another(self):
    self.assertEqual(True, R.lt(3, 5))
    self.assertEqual(False, R.lt(6, 5))
    self.assertEqual(False, R.lt(7.0, 7.0))
    self.assertEqual(True, R.lt('abc', 'xyz'))
    self.assertEqual(False, R.lt('abcd', 'abc'))


if __name__ == '__main__':
  unittest.main()



from .private._arity import _arity
from .private._curry1 import _curry1
from .private._curry2 import _curry2
from .private._curryN import _curryN

curryN = _curry2(lambda n, fn: _curry1(fn) if n == 1 else _arity(n, _curryN(n, [], fn)))



from ._helper import getAttribute
from ._xfBase import XfBase


class XTap(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f

  def step(self, result, _input):
    self.f(_input)
    return getAttribute(self.xf, '@@transducer/step')(result, _input)


def _xtap(f): return lambda xf: XTap(f, xf)



from ._helper import getAttribute


def _xArrayReduce(xf, acc, arr):
  idx = 0
  length = len(arr)
  while idx < length:
    acc = getAttribute(xf, '@@transducer/step')(acc, arr[idx])
    if acc and getAttribute(acc, '@@transducer/reduced'):
      acc = getAttribute(acc, '@@transducer/value')
      break
    idx += 1
  return getAttribute(xf, '@@transducer/result')(acc)



import copy

from .curryN import curryN
from .keys import keys
from .private._curry2 import _curry2
from .private._dispatchable import _dispatchable
from .private._has import _has
from .private._helper import funcArgsLength
from .private._isFunction import _isFunction
from .private._map import _map
from .private._reduce import _reduce
from .private._xmap import _xmap


def inner_map(fn, functor):
  def inner_reduce(acc, key):
    """
    There are 2 cases of functor
    case 1: functor is a dict or an instance with get method
    case 2: functor is an instance of some classes
    """
    if isinstance(functor, dict) or _has(functor, 'get'):
      acc[key] = fn(functor.get(key))
    else:
      setattr(acc, key, fn(getattr(acc, key, None)))
    return acc

  if functor is None:
    raise Exception('Can not work with None')
  if _isFunction(functor):
    return curryN(funcArgsLength(functor), lambda *arguments: fn(functor(*arguments)))
  if isinstance(functor, (list, tuple)):
    return _map(fn, functor)
  return _reduce(inner_reduce, {} if isinstance(functor, dict) or _has(functor, 'get') else copy.deepcopy(functor), keys(functor))


# pylint: disable=redefined-builtin
map = _curry2(_dispatchable(['fantasy-land/map', 'map'], _xmap, inner_map))



from ._helper import safeLen
from ._isArrayLike import _isArrayLike


def _makeFlat(recursive):
  def flatt(arr):
    result = []
    idx = 0
    ilen = len(arr)

    while idx < ilen:
      if _isArrayLike(arr[idx]):
        value = flatt(arr[idx]) if recursive else arr[idx]
        j = 0
        jlen = safeLen(value)
        while j < jlen:
          result.append(value[j])
          j += 1
      else:
        result.append(arr[idx])
      idx += 1
    return result
  return flatt




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/invoker.js
"""


concat2 = R.invoker(2, 'concat')


class Concater:
  def __init__(self, xs):
    self.xs = xs

  def concat(self, a, b):
    return self.xs + [a] + [b]


class TestInvoker(unittest.TestCase):
  def test_calls_the_method_on_the_object(self):
    self.assertEqual([1, 2, 3, 4], concat2(3, 4, Concater([1, 2])))

  def test_raises_a_descriptive_TypeError_if_method_does_not_exist(self):
    with self.assertRaises(TypeError):
      self.assertEqual([1, 2, 3, 4], concat2(3, 4, [1, 2]))

  def test_curries_the_method_call(self):
    self.assertEqual([1, 2, 3, 4], concat2(3)(4)(Concater([1, 2])))
    self.assertEqual([1, 2, 3, 4], concat2(3, 4)(Concater([1, 2])))
    self.assertEqual([1, 2, 3, 4], concat2(3)(4, Concater([1, 2])))


if __name__ == '__main__':
  unittest.main()



from .private._curry1 import _curry1
from .private._isString import _isString
from .private._quote import _quote


def inner_toString(val):
  """
  TODO: dict, set, list, tuple, function, object
  TODO: regex, date if needed
  """
  if _isString(val):
    return _quote(val)
  return str(val)


toString = _curry1(inner_toString)



from .clone import clone
from .private._curry1 import _curry1
from .private._helper import getAttribute
from .private._isArray import _isArray
from .private._isDict import _isDict
from .private._isFunction import _isFunction
from .private._isSet import _isSet
from .private._isString import _isString


def inner_empty(x):
  if x is not None and _isFunction(getAttribute(x, 'fantasy-land/empty')):
    return getAttribute(x, 'fantasy-land/empty')()
  if x is not None and _isFunction(getAttribute(x, 'empty')):
    return getAttribute(x, 'empty')()
  if _isArray(x):
    return []
  if _isString(x):
    return ''
  if _isDict(x):
    return {}
  if _isSet(x):
    return set()
  if x is not None:
    # by default, we will not empty other types for now.
    return clone(x)
  return None


empty = _curry1(inner_empty)



from .keys import keys
from .private._curry1 import _curry1
from .private._helper import getAttribute


def inner_values(obj):
  props = keys(obj)
  length = len(props)
  vals = []
  idx = 0
  while idx < length:
    val = getAttribute(obj, props[idx])
    vals.append(val)
    idx += 1
  return vals


values = _curry1(inner_values)



from .equals import equals
from .private._curry2 import _curry2
from .private._filter import _filter
from .private._Set import _Set
from .uniq import uniq


def inner_intersection(list1, list2):
  toKeep = _Set()

  for item in list1:
    toKeep.add(item)

  # Below: change to a shorter version
  return uniq(_filter(lambda item: len([x for x in toKeep if equals(x, item)]), list2))


intersection = _curry2(inner_intersection)



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/adjust.js
"""


class TestAdjust(unittest.TestCase):
  def test_applies_the_given_function_to_the_value_at_the_given_index_of_the_supplied_array(self):
    self.assertEqual([0, 1, 3, 3], R.adjust(2, R.add(1), [0, 1, 2, 3]))

  def test_offsets_negative_indexes_from_the_end_of_the_array(self):
    self.assertEqual([0, 2, 2, 3], R.adjust(-3, R.add(1), [0, 1, 2, 3]))

  def test_returns_the_original_array_if_the_supplied_index_is_out_of_bounds(self):
    arr = [0, 1, 2, 3]
    self.assertEqual(arr, R.adjust(4, R.add(1), arr))
    self.assertEqual(arr, R.adjust(-5, R.add(1), arr))

  def test_does_not_mutate_the_original_array(self):
    arr = [0, 1, 2, 3]
    self.assertEqual([0, 1, 3, 3], R.adjust(2, R.add(1), arr))
    self.assertEqual([0, 1, 2, 3], arr)

  def test_accepts_an_array_like_object(self):
    pass
    # TODO: find a way to implement this in python style


if __name__ == '__main__':
  unittest.main()




import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/isEmpty.js
"""


class TestIsEmpty(unittest.TestCase):
  def test_returns_False_for_None(self):
    self.assertEqual(False, R.isEmpty(None))

  def test_returns_True_for_empty_string(self):
    self.assertEqual(True, R.isEmpty(''))
    self.assertEqual(False, R.isEmpty(' '))

  def test_returns_True_for_empty_array(self):
    self.assertEqual(True, R.isEmpty([]))
    self.assertEqual(False, R.isEmpty([[]]))

  def test_returns_True_for_empty_dict(self):
    self.assertEqual(True, R.isEmpty({}))
    self.assertEqual(False, R.isEmpty({'x': 0}))

  def test_returns_True_for_empty_set(self):
    self.assertEqual(True, R.isEmpty(set()))
    self.assertEqual(False, R.isEmpty(set([1])))

  def test_object_is_not_empty(self):
    class Obj:
      pass
    self.assertEqual(False, R.isEmpty(Obj()))


if __name__ == '__main__':
  unittest.main()



from ..objOf import objOf
from ._identity import _identity
from ._isArrayLike import _isArrayLike
from ._isTransformer import _isTransformer


def _array_step(xs, x):
  return xs + [x]


_stepCatArray = {
    '@@transducer/init': list,
    '@@transducer/step': _array_step,
    '@@transducer/result': _identity
}
_stepCatString = {
    '@@transducer/init': str,
    '@@transducer/step': lambda a, b: str(a) + str(b),
    '@@transducer/result': _identity
}

_stepCatDict = {
    '@@transducer/init': dict,
    # use Python 3.9 feature
    '@@transducer/step': lambda result, input: result | (objOf(input[0], input[1]) if _isArrayLike(input) else input),
    '@@transducer/result': _identity
}

# TODO: add _stepCatObject


def _stepCat(obj):
  if _isTransformer(obj):
    return obj
  if _isArrayLike(obj):
    return _stepCatArray
  if isinstance(obj, str):
    return _stepCatString
  if isinstance(obj, dict):
    return _stepCatDict
  raise Exception(f'Cannot create transformer for {obj}')




import datetime
import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/max.js
"""


class TestMax(unittest.TestCase):
  def test_returns_the_larger_of_its_two_arguments(self):
    self.assertEqual(7, R.Max(-7, 7))
    self.assertEqual(7, R.Max(7, -7))

  def test_works_for_any_orderable_type(self):
    d1 = datetime.date(2001, 1, 1)
    d2 = datetime.date(2002, 2, 2)
    self.assertEqual(d2, R.Max(d1, d2))
    self.assertEqual(d2, R.Max(d2, d1))

    self.assertEqual('b', R.Max('a', 'b'))
    self.assertEqual('b', R.Max('b', 'a'))

  def test_returns_the_second_argument_if_both_arguments_are_equal(self):
    self.assertEqual(7, R.Max(7, 7))
    self.assertEqual(None, R.Max(None, None))

  def test_returns_the_alphabetically_later_type_if_neigher_argument_is_greater_than_the_other(self):
    self.assertEqual('a', R.Max('a', 7))
    self.assertEqual(None, R.Max('A', None))

  def test_returns_the_alphabetically_later_string_coersion_if_no_argument_or_type_is_greater_than_the_other(self):
    obj1 = {'a': 1}
    obj2 = {'a': 1}
    self.assertEqual(obj2, R.Max(obj1, obj2))
    self.assertEqual(obj1, R.Max(obj1, None))

  def test_returns_the_first_argument_if_no_other_comparison_attemps_produce_a_result(self):
    obj1 = {'a': 1}
    self.assertEqual(obj1, R.Max(obj1, obj1))
    self.assertTrue(R.equals(float('nan'), R.Max(float('nan'), float('nan'))))


if __name__ == '__main__':
  unittest.main()



import unittest

import ramda as R

from .helpers.Maybe import Just

"""
https://github.com/ramda/ramda/blob/master/test/indexOf.js
"""

input = [1, 2, 3, 4, 5]


class TestIndexOf(unittest.TestCase):
  def test_returns_a_number_indicating_an_object_position_in_a_list(self):
    arr = [0, 10, 20, 30]
    self.assertEqual(3, R.indexOf(30, arr))

  def test_returns_minus_one_if_the_object_is_not_in_the_list(self):
    arr = [0, 10, 20, 30]
    self.assertEqual(-1, R.indexOf(40, arr))

  def test_returns_the_index_of_the_first_item(self):
    self.assertEqual(0, R.indexOf(1, input))

  def test_returns_the_index_of_the_last_item(self):
    self.assertEqual(4, R.indexOf(5, input))

  def test_finds_1(self):
    self.assertEqual(0, R.indexOf(1, [1, 2, 3]))

  def test_does_not_consider_str_1_equal_to_1(self):
    self.assertEqual(-1, R.indexOf('1', [1, 2, 3]))

  def test_returns_minus_1_for_an_empty_array(self):
    self.assertEqual(-1, R.indexOf('x', []))

  def test_has_R_equals_semantics(self):
    # TODO: ignore neg-zero and pos-zero check for now, due to simlicity
    # self.assertEqual(1, len(R.difference([0], [-0])))
    self.assertEqual(0, R.indexOf(float('nan'), [float('nan')]))
    self.assertEqual(0, R.indexOf(Just([42]), [Just([42])]))

  def test_dispatches_to_indexOf_method(self):
    class Empty:
      def indexOf(self, *ignore):
        """
        indexOf method signature should be the same with List
        """
        return -1

    class List:
      def __init__(self, head, tail):
        self.head = head
        self.tail = tail

      def indexOf(self, x):
        idx = self.tail.indexOf(x)
        if self.head == x:
          return 0
        elif idx >= 0:
          return 1 + idx
        else:
          return -1

    arr = List('b',
               List('a',
                    List('n',
                         List('a',
                              List('n',
                                   List('a',
                                        Empty()
                                        )
                                   )
                              )
                         )
                    )
               )

    self.assertEqual(1, R.indexOf('a', 'banana'))
    self.assertEqual(-1, R.indexOf('x', 'banana'))
    self.assertEqual(1, R.indexOf('a', arr))
    self.assertEqual(-1, R.indexOf('x', arr))

  def test_finds_function_compared_by_identity(self):
    def f(): return None
    def g(): return None
    arr = [g, f, g, f]
    self.assertEqual(1, R.indexOf(f, arr))

  def test_does_not_find_function_compared_by_identity(self):
    def f(): return None
    def g(): return None
    def h(): return None
    arr = [g, f]
    self.assertEqual(-1, R.indexOf(h, arr))


if __name__ == '__main__':
  unittest.main()



from ._arrayReduce import _arrayReduce
from ._createReduce import _createReduce
from ._helper import getAttribute


def _iterableReduce(reducer, acc, _iter):

  while True:
    try:
      value = next(_iter)
      acc = reducer(acc, value)
    except StopIteration:
      break
  return acc


def _methodReduce(reducer, acc, obj, methodName):
  method = getAttribute(obj, methodName)
  return method(reducer, acc)


_reduce = _createReduce(_arrayReduce, _methodReduce, _iterableReduce)



from .private._curry3 import _curry3
from .private._helper import getAttribute
from .private._isTransformer import _isTransformer
from .private._stepCat import _stepCat
from .private._xReduce import _xReduce


def inner_into(acc, transducer, arr):
  xf = transducer(acc if _isTransformer(acc) else _stepCat(acc))
  return _xReduce(xf, getAttribute(xf, '@@transducer/init')(), arr)


into = _curry3(inner_into)



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/prop.js
"""

fred = {'name': 'Fred', 'age': 23}


class TestPropForDict(unittest.TestCase):
  def test_returns_a_function_that_fetches_the_appropriate_property(self):
    nm = R.prop('name')
    self.assertEqual('Fred', nm(fred))

  def test_handles_number_as_property(self):
    deities = ['Cthulhu', 'Dagon', 'Yog-Sothoth']
    self.assertEqual('Cthulhu', R.prop(0, deities))
    self.assertEqual('Dagon', R.prop(1, deities))
    self.assertEqual('Yog-Sothoth', R.prop(2, deities))
    self.assertEqual('Yog-Sothoth', R.prop(-1, deities))

  def test_shows_the_same_behaviour_as_path_for_a_nonexistent_property(self):
    propResult = R.prop('incorrect', fred)
    pathResult = R.path(['incorrect'], fred)
    self.assertEqual(propResult, pathResult)

  def test_shows_the_same_behaviour_as_path_for_a_None_property(self):
    # TODO: fix getAttribute
    # propResult = R.prop(None, fred)
    # pathResult = R.path([None], fred)
    # self.assertEqual(propResult, pathResult)
    pass

  def test_shows_the_same_behaviour_as_path_for_a_valid_property_and_object(self):
    propResult = R.prop('age', fred)
    pathResult = R.path(['age'], fred)
    self.assertEqual(propResult, pathResult)

  def test_shows_the_same_behaviour_as_path_for_a_None_object(self):
    propResult = R.prop('age', None)
    pathResult = R.path(['age'], None)
    self.assertEqual(propResult, pathResult)

class Fred:
  def __init__(self):
    self.name = 'Fred'
    self.age = 23
fredObj = Fred()
class TestPropForObject(unittest.TestCase):
  def test_returns_a_function_that_fetches_the_appropriate_property(self):
    nm = R.prop('name')
    self.assertEqual('Fred', nm(fredObj))

  def test_shows_the_same_behaviour_as_path_for_a_nonexistent_property(self):
    propResult = R.prop('incorrect', fredObj)
    pathResult = R.path(['incorrect'], fredObj)
    self.assertEqual(propResult, pathResult)

  def test_shows_the_same_behaviour_as_path_for_a_None_property(self):
    # TODO: fix getAttribute
    # propResult = R.prop(None, fredObj)
    # pathResult = R.path([None], fredObj)
    # self.assertEqual(propResult, pathResult)
    pass

  def test_shows_the_same_behaviour_as_path_for_a_valid_property_and_object(self):
    propResult = R.prop('age', fredObj)
    pathResult = R.path(['age'], fredObj)
    self.assertEqual(propResult, pathResult)

  def test_shows_the_same_behaviour_as_path_for_a_None_object(self):
    propResult = R.prop('age', None)
    pathResult = R.path(['age'], None)
    self.assertEqual(propResult, pathResult)

if __name__ == '__main__':
  unittest.main()



from .private._curry2 import _curry2
from .private._helper import getAttribute
from .private._isArray import _isArray
from .private._isFunction import _isFunction
from .private._isString import _isString
from .toString import toString


def inner_concat(a, b):
  if _isArray(a):
    if _isArray(b):
      return a + b
    raise Exception(f"{toString(b)} is not an array")
  if _isString(a):
    if _isString(b):
      return a + b
    raise Exception(f"{toString(b)} is not a string")
  if a is not None and _isFunction(getAttribute(a, 'fantasy-land/concat')):
    return a.get('fantasy-land/concat')(b)
  if a is not None and _isFunction(a.concat):
    return a.concat(b)
  raise Exception(f'{a} does not have a method named "concat" or "fantasy-land/concat"')


concat = _curry2(inner_concat)




import unittest
from datetime import date
from math import isnan

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/subtract.js
"""


class TestSubtract(unittest.TestCase):
  def test_subtracts_two_numbers(self):
    self.assertEqual(15, R.subtract(22, 7))

  def test_coerces_its_arguments_to_numbers(self):
    self.assertEqual(-1, R.subtract('1', '2'))
    self.assertEqual(-1, R.subtract(1, '2'))
    self.assertEqual(1, R.subtract(True, False))
    self.assertTrue(isnan(R.subtract(None, None)))
    self.assertTrue(isnan(R.subtract(date(1, 1, 1), date(2, 1, 1))))


if __name__ == '__main__':
  unittest.main()



from ._helper import getAttribute
from ._reduced import _reduced
from ._xfBase import XfBase


class XAll(XfBase):
  def __init__(self, f, xf):
    self.xf = xf
    self.f = f
    self.all = True

  def result(self, result):
    if self.all:
      result = getAttribute(self.xf, '@@transducer/step')(result, True)
    return self.xf.get('@@transducer/result')(result)

  def step(self, result, _input):
    if not self.f(_input):
      self.all = False
      result = _reduced(getAttribute(self.xf, '@@transducer/step')(result, False))
    return result


def _xall(f): return lambda xf: XAll(f, xf)



import unittest

import ramda as R

"""
https://github.com/ramda/ramda/blob/master/test/identity.js
"""


class TestIdentity(unittest.TestCase):
  def test_returns_its_first_argument(self):
    self.assertEqual(None, R.identity(None))
    self.assertEqual('foo', R.identity('foo'))
    self.assertEqual('foo', R.identity('foo', 'bar'))


if __name__ == '__main__':
  unittest.main()
